""" Runs an python module in a separate process.
  Usage:
    runner = Runner(module_path)
    runner.execute(list-of-args, stdin-string)
    if runner.return_code == 0:
      result = runner.output  # strings written to stdout
"""


import os.path
import subprocess


class Runner(object):
  """Runs a python module as a program."""
  
  def __init__(self, python_module):
    self._python_module = python_module
    self.return_code = None  # Return code from running process
    self.output= None  # Lines written to stdout

  def execute(self, args, stdin_str):
    """
    Runs the child process from the parent process.
    Runs in the process of the caller.
    :param list-str args: list of arguments passed
       may be []
    :param str stdin_str: input to the program passed to stdin
    """
    arg_list = ['python', self._python_module]
    arg_list.extend(args)
    process = subprocess.run(
        arg_list,
        stdout=subprocess.PIPE,
        input=stdin_str, encoding='ascii')
    self.return_code = process.returncode
    self.output = process.stdout



# -*- coding: utf-8 -*-
"""Doctrings on print_model.py.

This script has a simple function named print_model,
which prints out all reactions in a given model.
The format is as follows:
'reaction' 'reaction index': 'reactants' '->' 'products'

The goal is to help examine models, for example
to manually verify whether a model is stoichiometrically inconsistent.

"""



def print_model(model):
    """Print reactants and products of all reactions in a model.


    Args:
        model (sbml_model): An sbml model to print out

    Returns:
        None: Just prints out the reactions
        
    """

    # print model name
    print(model)

    # run for each reaction
    for reaction_idx in range(model.getNumReactions()):
        reaction = model.getReaction(reaction_idx)

        #stg = "reaction " + str(reaction_idx+1) + ": "
        stg = reaction.getId() + ": "

        reactants_list = [reactant.getSpecies() for reactant in reaction.getListOfReactants()]
        products_list = [product.getSpecies() for product in reaction.getListOfProducts()]
        stg = stg + ' + '.join(reactants_list) + " -> " + ' + '.join(products_list) + ';'
        
        print(stg)



from SBMLLint.common import config
from SBMLLint.common import constants as cn

import os
import unittest


IGNORE_TEST = True
TEST_BAD_CONFIG_FILE = os.path.join(cn.TEST_DIR,
    "test_cfg_bad.yml")


#############################
# Tests
#############################
class TestFunctions(unittest.TestCase):

  def tearDown(self):
    config.setConfiguration()

  def verifyDefaultConfiguration(self):
    result = config.getConfiguration()
    self.assertTrue(isinstance(result, dict))
    for k, v in cn.CFG_DEFAULTS.items():
      self.assertTrue(k in result)
      self.assertEqual(result[k], cn.CFG_DEFAULTS[k])

  def testGetConfiguration(self):
    if IGNORE_TEST:
      return
    self.verifyDefaultConfiguration()

  def testSetConfiguration(self):
    if IGNORE_TEST:
      return
    config._config_dict = {}
    self.assertEqual(len(config.getConfiguration()), 0)
    config.setConfiguration()
    self.verifyDefaultConfiguration()

  def testSetConfiguration2(self):
    # TESTING
    with self.assertRaises(SystemExit):
      config.setConfiguration(path=TEST_BAD_CONFIG_FILE)
      
    

if __name__ == '__main__':
  unittest.main()



"""Chemical Reaction."""

from SBMLLint.common import constants as cn
from SBMLLint.common.molecule import Molecule, MoleculeStoichiometry

import numpy as np


REACTION_SEPARATOR = "->"


################# FUNCTIONS ###################
def getMolecules(libsbml_reaction, func):
  """
  Constructs molecules for the species returned by function.
  :param Function func: gets libsbml.SpeciesReference
  :return list-MoleculeStoichiometry:
  """
  species = func(libsbml_reaction)
  molecule_stoichiometrys = []
  for spc in species:
    molecule = Molecule.getMolecule(spc.getSpecies())
    if molecule is None:
        molecule = Molecule(spc.getSpecies())
    molecule_stoichiometrys.append(MoleculeStoichiometry(
        molecule,
        spc.getStoichiometry())
        )
  return molecule_stoichiometrys


################# Classes ###################
class Reaction(object):

  def __init__(self, libsbml_reaction):
    self.reactants = self.makeMoleculeStoichiometrys(
        libsbml_reaction.getReactant,
        libsbml_reaction.getNumReactants)
    self.products = self.makeMoleculeStoichiometrys(
        libsbml_reaction.getProduct,
        libsbml_reaction.getNumProducts)
    if libsbml_reaction.getKineticLaw() is not None:
      self.kinetics_law = libsbml_reaction.getKineticLaw().getFormula()
    else:
      self.kinetics_law = None
    self.label = libsbml_reaction.getId()
    self.identifier = self.makeIdentifier(is_include_kinetics=True)
    self.category = self.getCategory()
    self.kinetics_terms = self.getKineticsTerms(libsbml_reaction)

  def makeMoleculeStoichiometrys(self, func_get_one, func_get_num):
    """
    Creates a list of MoleculeStoichiometry
    :param Function func_get_one: get one element by index
    :param Function func_get_num: get number of elements
    :return list-MoleculeStoichiometry:
    """
    result = []
    collection = [func_get_one(n) for n in range(func_get_num())]
    for s_r in collection:
      molecule = Molecule(s_r.getSpecies())
      stoich = s_r.getStoichiometry()
      result.append(MoleculeStoichiometry(molecule, stoich))
    return result

  def getId(self, is_include_kinetics=True, is_include_label=True):
    """
    Constructs an ID that may be a substring
    of the the full reaction identifier.
    :param bool is_include_kinetics: Include the kinetics law
    :return str:
    """
    result = self.identifier
    if not is_include_kinetics:
      pos = result.find(cn.KINETICS_SEPARATOR)
      if pos > 0:
        result = result[:pos]
    if not is_include_label:
      pos = result.find(cn.LABEL_SEPARATOR)
      if pos > 0:
        result = result[pos+2:]  # Eliminate the separator and space
    return result

  def makeIdentifier(self, is_include_kinetics=True):
    """
    Provides a string representation of the reaction
    :param bool is_include_kinetics: include the kinetics formula
    :return str:
    """
    def makeStoichiometryString(molecule_stoichiometry):
      num = molecule_stoichiometry.stoichiometry
      if np.isclose(num, 1.0):
        return ''
      else:
        return "%2.2f " % num
    #
    def makeTermCollection(molecule_stoichiometries):
      """
      Formats a set of terms with stoichiometries.
      :param list-MoleculeStoichiometry:
      :return str:
      """
      term_collection = ''
      for m_s in molecule_stoichiometries:
        term = "%s%s" % (makeStoichiometryString(m_s), str(m_s.molecule))
        if len(term_collection) == 0:
          term_collection += term
        else:
          term_collection += " + " + term
      return term_collection
    #
    reactant_collection = makeTermCollection(self.reactants)
    product_collection = makeTermCollection(self.products)
    if is_include_kinetics:
      if self.kinetics_law is None:
        formula_str = ''
      else:
        formula_str = "; " + self.kinetics_law
    else:
      formula_str = ''
    reaction_str = "%s: %s -> %s" % (self.label,
        reactant_collection, product_collection)
    reaction_str = reaction_str + formula_str
    return reaction_str

  def __repr__(self):
    return self.identifier

  def getCategory(self):
    """
    :return str: reaction category
    """
    num_reactants = len([r.molecule for r in self.reactants \
                         if r.molecule.name!=cn.EMPTYSET])
    num_products = len([p.molecule for p in self.products \
                        if p.molecule.name!=cn.EMPTYSET])
    stoichiometry_reactants = sum([r.stoichiometry for r \
                                   in self.reactants \
                                   if r.molecule.name!=cn.EMPTYSET])
    stoichiometry_products = sum([p.stoichiometry for p \
                                  in self.products \
                                  if p.molecule.name!=cn.EMPTYSET])
    for reaction_category in cn.REACTION_CATEGORIES:
      if reaction_category.predicate(num_reactants, num_products, 
                                     stoichiometry_reactants, 
                                     stoichiometry_products):
        return reaction_category.category
    raise ValueError("Reaction category not found.")

  def isEqual(self, other_reaction):
    """
    Checks if two reactions are the same.
    :param Reaction other_reaction:
    :return bool:
    """
    return self.identifier == other_reaction.identifier

  @classmethod
  def initialize(cls, simple):
    """
    :param SimpleSBML simple:
    """
    cls.reactions = []
    [Reaction(r) for r in simple.reactions]

  def getKineticsTerms(self, libsbml_reaction):
    """
    Gets the terms used in the kinetics law for the reaction
    :param libsbml.Reaction libsbml_reaction:
    :return list-of-str: names of the terms
    """
    terms = []
    law = libsbml_reaction.getKineticLaw()
    if law is not None:
      math = law.getMath()
      asts = [math]
      while len(asts) > 0:
        this_ast = asts.pop()
        if this_ast.isName():
          terms.append(this_ast.getName())
        else:
          pass
        num = this_ast.getNumChildren()
        for idx in range(num):
          asts.append(this_ast.getChild(idx))
    return terms

  @classmethod
  def find(cls, reactions, category=cn.REACTION_1_1):
    return [r for r in reactions if r.category == category]
  



from SBMLLint.common import constants as cn
from SBMLLint.tools import model_maker

import os
import unittest


IGNORE_TEST = False
TEST_IN_FILE = "test_model_maker.txt"
TEST_IN_FILE2 = "test_model_maker2.txt"
TEST_IN_PATH = os.path.join(cn.TEST_DIR, TEST_IN_FILE)
TEST_IN_PATH2 = os.path.join(cn.TEST_DIR, TEST_IN_FILE2)
TEST_OUT_FILE = "test_model_maker.ant"
TEST_OUT_PATH = os.path.join(cn.TEST_DIR, TEST_OUT_FILE)
SYM1 = "sym1"
SYM2 = "sym2"
SYM3 = "sym3"
SYM4 = "sym4"
REACTIONSTR = "%s + %s -> %s + %s" % (SYM1, SYM2, SYM3, SYM4)


#############################
# Tests
#############################
class TestModelMaker(unittest.TestCase):

  def tearDown(self):
    if os.path.isfile(TEST_OUT_PATH):
      os.remove(TEST_OUT_PATH)

  def setUp(self):
    self.maker = model_maker.ModelMaker(REACTIONSTR)

  def testConstructor(self):
    if IGNORE_TEST:
      return
    self.assertEqual(self.maker._reaction_strs, [REACTIONSTR])
    self.maker = model_maker.ModelMaker([REACTIONSTR])
    self.assertEqual(self.maker._reaction_strs, [REACTIONSTR])
    self.maker = model_maker.ModelMaker(TEST_IN_PATH)
    self.assertGreater(len(self.maker._reaction_strs), 0)

  def testExtractSymbols(self):
    if IGNORE_TEST:
      return
    symbols = self.maker.extractSymbols()
    b = set(symbols).symmetric_difference([SYM1, SYM2, SYM3, SYM4])
    self.assertEqual(len(b), 0)
    self.maker = model_maker.ModelMaker([REACTIONSTR, REACTIONSTR])
    b = set(symbols).symmetric_difference([SYM1, SYM2, SYM3, SYM4])
    self.assertEqual(len(b), 0)

  def testMakeModelStr(self):
    if IGNORE_TEST:
      return
    model_str = self.maker.makeModelStr()
    self.assertGreater(len(model_str), len(REACTIONSTR))
   
  def testMakeModelFile(self): 
    if IGNORE_TEST:
      return
    self.assertFalse(os.path.isfile(TEST_OUT_PATH))
    model_str = self.maker.makeModelFile(TEST_OUT_PATH)
    self.assertTrue(os.path.isfile(TEST_OUT_PATH))

  def testSplitNumber(self):
    if IGNORE_TEST:
      return
    stg = "ab"
    a_str, a_int = self.maker.__class__._splitNumber(stg)
    self.assertEqual(a_str, stg)
    self.assertIsNone(a_int)
    #
    prefix = "ab"
    num = 1
    stg = "%s%d" % (prefix, num)
    a_str, a_int = self.maker.__class__._splitNumber(stg)
    self.assertEqual(a_str, prefix)
    self.assertEqual(a_int, num)
    #
    prefix = "ab"
    num = 12
    stg = "%s%d" % (prefix, num)
    a_str, a_int = self.maker.__class__._splitNumber(stg)
    self.assertEqual(a_str, prefix)
    self.assertEqual(a_int, num)
    #
    prefix = "a1b"
    num = 12
    stg = "%s%d" % (prefix, num)
    a_str, a_int = self.maker.__class__._splitNumber(stg)
    self.assertEqual(a_str, prefix)
    self.assertEqual(a_int, num)

  def testRepetitionName(self):
    if IGNORE_TEST:
      return
    def test(rname, expected, exclude_funcs=None):
      result = self.maker.__class__._makeRepetitionNames(rname,
          exclude_funcs=exclude_funcs)
      self.assertEqual(result, expected)
    #
    test("ab__bB12", "ab__bB12", exclude_funcs=[lambda n: n[0]=="b"])
    test("ab12", "ab_12")
    test("a__bb12", "a__bb_12")
    test("ab__bB12", "ab__bB_12")
    test("ab__bB12", "ab__bB_12", exclude_funcs=[lambda n: n[0]=="a"])
    test("E3SUB__SUB__misfolded__Ub2__UCHL1",
        "E3SUB__SUB__misfolded__Ub_2__UCHL1",
        exclude_funcs=[lambda n: "UCHL" in n])

  def testGetCandidateRenames(self):
    if IGNORE_TEST:
      return
    candidates = ["a_b12", "a1"]
    non_candidates = ["a1b", "a23c"]
    self.maker.symbols = list(candidates)
    self.maker.symbols.extend(non_candidates)
    result = self.maker.getCandidateRenames()
    self.assertEqual(len(candidates), len(result))

  def testGetCandidateRenames(self):
    if IGNORE_TEST:
      return
    maker = model_maker.ModelMaker(TEST_IN_PATH2)
    rename_dict1 = maker.getCandidateRenames()
    exclude_funcs = [
        lambda n: n[0]=="k",
        lambda n: n=="E1",
        lambda n: n=="E2",
        lambda n: n=="E3",
        lambda n: "UCHL1" in n,
        lambda n: "Uchl1" in n,
        lambda n: n[-2:] == "E3",
    ]
    rename_dict2 = maker.getCandidateRenames(
        exclude_funcs=exclude_funcs)
    self.assertGreater(len(rename_dict1), len(rename_dict2))

  def testReplaceSymbols(self):
    if IGNORE_TEST:
      return
    model_str = self.maker.makeModelStr()
    self.maker.replaceSymbols({SYM1: SYM2, SYM3: SYM4})
    for sym in [SYM2, SYM4]:
      self.assertEqual(self.maker.model_str.count(sym), 4)
    for sym in [SYM1, SYM3]:
      self.assertEqual(self.maker.model_str.count(sym), 0)


if __name__ == '__main__':
  unittest.main()



"""
Tests for Moiety and MoietyStoichiometry
"""
from SBMLLint.common import constants as cn
from SBMLLint.common import config
from SBMLLint.common.moiety import Moiety, MoietyStoichiometry
from SBMLLint.common import util

import itertools
import numpy as np
import os
import unittest


IGNORE_TEST = False
MOIETY_NAME1 = "first"
MOIETY_NAME2 = "second"
NUM1 = 2
MOIETY_STOICHIOMETRY_STGS = {
    ("P", 1): ["P", "P_1"],
    ("PP", 2): ["PP_2"],
    }
MOIETY_STRUCTURE =  [{'ATP': ['A, 1', 'P, 3']},
    {'Glu6P': ['Glu, 1', 'P, 1']}, {'ADP': ['A, 1', 'P, 2']}]
TEST_CFG_FILE = os.path.join(cn.TEST_DIR, "test_sbmllint_cfg.yml")

#######################################
class TestMoiety(unittest.TestCase):

  def testConstructor(self):
    if IGNORE_TEST:
      return
    self.assertEqual(Moiety(MOIETY_NAME1).name, MOIETY_NAME1)


class TestMoietyStoichiometry(unittest.TestCase):

  def testStoichiometry(self):
    moiety_stoichiometry = MoietyStoichiometry(Moiety(MOIETY_NAME1),
        NUM1)
    self.assertEqual(moiety_stoichiometry.moiety.name,
        MOIETY_NAME1)
    self.assertEqual(moiety_stoichiometry.stoichiometry, NUM1)
 
  def testMake(self):
    if IGNORE_TEST:
      return
    for expected, strings in MOIETY_STOICHIOMETRY_STGS.items():
      for stg in strings:
        result = MoietyStoichiometry.make(stg)
        self.assertEqual(result.moiety.name, expected[0])
        self.assertEqual(result.stoichiometry, expected[1])

  def testGetMoietys(self):
    if IGNORE_TEST:
      return
    m_ss = []
    for stgs in MOIETY_STOICHIOMETRY_STGS.values():
      m_ss.extend([MoietyStoichiometry.make(s) for s in stgs])
    moietys = MoietyStoichiometry.getMoietys(m_ss)
    self.assertEqual(len(MOIETY_STOICHIOMETRY_STGS.keys()),
        len(moietys))

  def testMakeFromDct(self):
    if IGNORE_TEST:
      return
    config.setConfiguration(TEST_CFG_FILE)
    config_dct = config.getConfiguration()
    dct = config_dct[cn.CFG_MOIETY_STRUCTURE]
    mss1 = MoietyStoichiometry.makeFromDct(dct["ATP"])
    mss2 = [MoietyStoichiometry("A", 1),
        MoietyStoichiometry("P", 3)]
    self.assertTrue(all([m1.isEqual(m2)
         for m1, m2 in zip(mss1, mss2)]))


if __name__ == '__main__':
  unittest.main()



#!/usr/bin/env python
"""
Runs moiety analysis for a local XML file.
Usage: moiety_analysis <filepath>
"""

from SBMLLint.common import constants as cn
from SBMLLint.common import util
from SBMLLint.tools import sbmllint

import argparse

def main():
  parser = argparse.ArgumentParser(description='SBML XML file.')
  parser.add_argument('xml_file', type=open, help='SBML or zip file')
  parser.add_argument('--config', type=open,
      help="SBMLLint configuration file")
  args = parser.parse_args()
  for fid in util.getNextFid(args.xml_file):
    util.runFunction(sbmllint.lint, kwargs={
        "model_reference": fid,
        "mass_balance_check": cn.MOIETY_ANALYSIS,
        "config_fid": args.config,
        })


if __name__ == '__main__':
  main()



"""
Tests for Mass Equality Set Graph (MESGraph)
"""
from SBMLLint.common import constants as cn
from SBMLLint.common.reaction import Reaction
from SBMLLint.common.simple_sbml import SimpleSBML
from SBMLLint.games.mesgraph import MESGraph
from SBMLLint.games.som import SOM
from SBMLLint.common import simple_sbml

import numpy as np
import os
import re
import unittest

IGNORE_TEST = False
# Constants for simple and simple2, non multi-multi reactions
INITIAL_NODES = 14
INITIAL_EDGES = 0
FINAL_NODES = 6
FINAL_EDGES = 6
UNIUNI0 = 0
UNIUNI1 = 7
UNIUNI2 = 8
UNIMULTI = 2
MULTIUNI = 13
INEQUAL1 = 14
INEQUAL2 = 15
# Constants for simple
AA = "AA"
CN = "Cn"
DFG = "DFG"
E1 = "E1"
E2 = "E2"
FRU = "Fru"
GLY = "Gly"
MEL = "Mel"
MG = "MG"
# Constants for simple2
REACTION1 = "SHMTr"
REACTION2 = "MTHFR"
REACTION3 = "MTR"
# Constants for simple3, Type III errors
V1 = "v1"
V2 = "v2"
V3 = "v3"
AMP = "AMP"
ADP = "ADP"
ATP = "ATP"
# Constants for simple4, Tyep IV errors, error from reducing
ATPASE = "ATPase"
LOWER = "lower"
FRU16P2 = "Fru16P2"
#simple5, type V error, SOM cycle
DIH = "DIH"
DIN = "DIN"
#simple6, no errors in multi-multi model
R5 = "R5"
R12 = "R12"
AC = "Ac"
ACP = "AcP"


#############################
# Tests
#############################
class TestMESGraph(unittest.TestCase):

  def setUp(self):
    # SimpleSBML with type I error
    self.simple = SimpleSBML()
    self.simple.initialize(cn.TEST_FILE6)
    # SimpleSBML with type II error
    self.simple2 = SimpleSBML()
    self.simple2.initialize(cn.TEST_FILE7)
    # SimpleSBML for type III error
    self.simple3 = SimpleSBML()
    self.simple3.initialize(cn.TEST_FILE10)
    # SimpleSBML for type IV error
    self.simple4 = SimpleSBML()
    self.simple4.initialize(cn.TEST_FILE11)
    # simple5 for type V error    
    self.simple5 = SimpleSBML()
    self.simple5.initialize(cn.TEST_FILE12)
    # simple6 will test multi-multi model with no errors
    self.simple6 = SimpleSBML()
    self.simple6.initialize(cn.TEST_FILE3)
    self.mesgraph = MESGraph(self.simple)
  
  def testConstructor(self):
    if IGNORE_TEST:
      return
    self.assertEqual(len(self.mesgraph.nodes), INITIAL_NODES)
    self.assertEqual(len(self.mesgraph.edges), INITIAL_EDGES)
    dfg = self.simple.getMolecule(DFG)
    # molecules is a list of one-molecule sets
    molecules = [som.molecules for som in self.mesgraph.nodes]
    self.assertTrue({self.simple.getMolecule(DFG)} in molecules)
    self.assertEqual(len(self.mesgraph.type_one_errors), 0)
    self.assertEqual(len(self.mesgraph.type_two_errors), 0)
    self.assertEqual(len(self.mesgraph.type_three_errors), 0)
    self.assertEqual(len(self.mesgraph.type_four_errors), 0)
    self.assertEqual(len(self.mesgraph.type_five_errors), 0)

  def testInitializeSOMs(self):
    if IGNORE_TEST:
      return
    for node in self.mesgraph.nodes:
      self.assertEqual(type(node), SOM)
  
  def testMakeId(self):
    if IGNORE_TEST:
      return
    identifier = ""
    for key, som in enumerate(self.mesgraph.nodes):
      identifier = identifier + som.identifier
      if key < len(self.mesgraph.nodes)-1:
        identifier = identifier + ";"
    self.assertEqual(identifier, self.mesgraph.identifier)

  def testGetNode(self):
    if IGNORE_TEST:
      return
    aa = self.simple.getMolecule(AA)
    aa_node = self.mesgraph.getNode(aa)
    self.assertEqual(type(aa_node), SOM)
    self.assertEqual(aa_node.molecules, {aa})
  
  def testMergeNodes(self):
    if IGNORE_TEST:
      return
    m3 = MESGraph(self.simple3)
    v1_reaction = self.simple3.getReaction(V1)
    v2_reaction = self.simple3.getReaction(V2)
    amp_molecule = self.simple3.getMolecule(AMP)
    adp_molecule = self.simple3.getMolecule(ADP)
    atp_molecule = self.simple3.getMolecule(ATP)
    amp_som = m3.getNode(amp_molecule)
    adp_som = m3.getNode(adp_molecule)
    atp_som = m3.getNode(atp_molecule)
    ampatp_som = m3.mergeNodes(amp_som, atp_som, v1_reaction)
    self.assertTrue(amp_molecule in ampatp_som.molecules)
    self.assertTrue(atp_molecule in ampatp_som.molecules)
    self.assertTrue(ampatp_som in m3.nodes)
    # tests if anohter merge will remove previous nodes
    ampadpatp_som = m3.mergeNodes(ampatp_som, adp_som, v2_reaction)
    self.assertFalse(ampatp_som in m3.nodes)
    self.assertFalse(adp_som in m3.nodes)  
    self.assertTrue(ampadpatp_som in m3.nodes)

  def testProcessUniUniReaction(self):
    if IGNORE_TEST:
      return
    self.mesgraph.processUniUniReaction(
        self.simple.reactions[UNIUNI0])
    dfg = self.mesgraph.getNode(self.simple.getMolecule(DFG))
    e1 = self.mesgraph.getNode(self.simple.getMolecule(E1))
    self.assertTrue(self.mesgraph.has_node(dfg))
    self.assertTrue(self.mesgraph.has_node(e1))
    self.assertEqual(dfg, e1)
  
  def testProcessUniMultiReaction(self):
    if IGNORE_TEST:
      return
    unimulti_reaction = self.simple.reactions[UNIMULTI]
    self.mesgraph.processUniMultiReaction(unimulti_reaction)
    prods = [self.mesgraph.getNode(product.molecule) for product in unimulti_reaction.products]
    dfg = self.mesgraph.getNode(self.simple.getMolecule(DFG))
    for prod in prods:
      self.assertTrue(self.mesgraph.has_edge(prod, dfg))
  
  def testProcessMultiUniReaction(self):
    if IGNORE_TEST:
      return
    multiuni_reaction = self.simple.reactions[MULTIUNI]
    self.mesgraph.processMultiUniReaction(multiuni_reaction)
    reacts = [self.mesgraph.getNode(reactant.molecule) for reactant in multiuni_reaction.reactants]
    mel = self.mesgraph.getNode(self.simple.getMolecule(MEL))
    for react in reacts:
      self.assertTrue(self.mesgraph.has_edge(react, mel))
  
  def testAddMultiMultiReaction(self):
    if IGNORE_TEST:
      return
    m3 = MESGraph(self.simple3)
    v2 = self.simple3.getReaction(V2)
    m3.addMultiMultiReaction(v2)
    self.assertTrue(v2 in m3.multimulti_reactions)

  def testAddTypeThreeError(self):
    if IGNORE_TEST:
      return
    m3 = MESGraph(self.simple3)
    v1 = self.simple3.getReaction(V1)
    v2 = self.simple3.getReaction(V2)
    v3 = self.simple3.getReaction(V3)
    m3.processUniUniReaction(v1)
    m3.processUniMultiReaction(v3)
    amp = m3.getNode(self.simple3.getMolecule(AMP))
    atp = m3.getNode(self.simple3.getMolecule(ATP))
    self.assertTrue(m3.addTypeThreeError(amp, atp, v2))
    self.assertTrue(len(m3.type_three_errors), 1)
    error = m3.type_three_errors[0]
    self.assertEqual(error.node1, amp)
    self.assertEqual(error.node2, atp)
    self.assertEqual(error.reactions, [v2.label])

  def testCheckTypeThreeError(self):
    if IGNORE_TEST:
      return
    m3 = MESGraph(self.simple3)
    v1 = self.simple3.getReaction(V1)
    v2 = self.simple3.getReaction(V2)
    v3 = self.simple3.getReaction(V3)
    m3.processUniUniReaction(v1)
    adp = m3.getNode(self.simple3.getMolecule(ADP))
    atp = m3.getNode(self.simple3.getMolecule(ATP))
    self.assertFalse(m3.checkTypeThreeError(adp, atp, v3))
    m3.processUniMultiReaction(v3) 
    self.assertTrue(m3.checkTypeThreeError(adp, atp, v2))
    self.assertTrue(m3.checkTypeThreeError(atp, adp, v2))

  def testReduceReaction(self):
    if IGNORE_TEST:
      return
    m4 = MESGraph(self.simple4)
    atpase = m4.simple.getReaction(ATPASE)
    lower = m4.simple.getReaction(LOWER)
    self.assertFalse(m4.reduceReaction(atpase))
    m4.processUniUniReaction(atpase)
    reduced_reaction = m4.reduceReaction(lower)
    self.assertEqual(type(reduced_reaction), Reaction)
    self.assertEqual(len(reduced_reaction.reactants), 1)
    self.assertEqual(reduced_reaction.products, [])   
    self.assertEqual(reduced_reaction.reactants[0].molecule.name, FRU16P2)

  def testProcessMultiMultiReactions(self):
    if IGNORE_TEST:
      return
    # type III error
    m3 = MESGraph(self.simple3)
    v1 = self.simple3.getReaction(V1)
    v2 = self.simple3.getReaction(V2)
    v3 = self.simple3.getReaction(V3)
    m3.processUniUniReaction(v1)
    m3.processUniMultiReaction(v3) 
    self.assertEqual(m3.type_three_errors, [])
    self.assertTrue(m3.processMultiMultiReaction(v2))
    self.assertEqual(len(m3.type_three_errors), 1) 
    # type IV error
    m4 = MESGraph(self.simple4)
    atpase = m4.simple.getReaction(ATPASE)
    lower = m4.simple.getReaction(LOWER)
    m4.processUniUniReaction(atpase)  
    self.assertEqual(m4.type_four_errors, []) 
    self.assertTrue(m4.processMultiMultiReaction(lower))
    self.assertEqual(len(m4.type_four_errors), 1)
    # no error
    m6 = MESGraph(self.simple6)
    r5 = m6.simple.getReaction(R5)
    r12 = m6.simple.getReaction(R12)
    m6.processUniUniReaction(r12)
    ac = m6.getNode(self.simple6.getMolecule(AC))
    acp = m6.getNode(self.simple6.getMolecule(ACP))
    self.assertFalse(ac==acp)
    self.assertTrue(m6.processMultiMultiReaction(r5))
    ac = m6.getNode(self.simple6.getMolecule(AC))
    acp = m6.getNode(self.simple6.getMolecule(ACP))
    self.assertTrue(ac==acp)

  def testAddArc(self):
    if IGNORE_TEST:
      return
    source = [self.mesgraph.getNode(self.simple.getMolecule(FRU)), 
              self.mesgraph.getNode(self.simple.getMolecule(GLY))]
    destination = [self.mesgraph.getNode(self.simple.getMolecule(E2))]
    dummy_reaction = self.simple.reactions[INEQUAL2]
    self.mesgraph.addArc(source[0], destination[0], dummy_reaction)
    self.mesgraph.addArc(source[1], destination[0], dummy_reaction)
    arc1 = [source[0], destination[0]]
    arc2 = [source[1], destination[0]]
    self.assertTrue(self.mesgraph.has_edge(arc1[0], arc1[1]))
    self.assertTrue(self.mesgraph.has_edge(arc2[0], arc2[1]))
    reaction_label1 = self.mesgraph.get_edge_data(arc1[0], arc1[1])[cn.REACTION][0]
    reaction_label2 = self.mesgraph.get_edge_data(arc2[0], arc1[1])[cn.REACTION][0]
    self.assertEqual(reaction_label1, dummy_reaction.label)
    self.assertEqual(reaction_label1, reaction_label2)
  
  def testGetSOMPath(self):
    if IGNORE_TEST:
      return
    uniuni_reaction = self.simple.reactions[UNIUNI0]
    self.mesgraph.processUniUniReaction(uniuni_reaction)
    dfg = self.simple.getMolecule(DFG)
    e1 = self.simple.getMolecule(E1)
    som = self.mesgraph.getNode(dfg)
    som_path = self.mesgraph.getSOMPath(som, e1, dfg)
    self.assertEqual(type(som_path[0]), cn.PathComponents)
    self.assertEqual(som_path[0].reactions, [uniuni_reaction.label])
    self.assertEqual(len(som_path), 1)

  def testPrintSOMPath(self):
    if IGNORE_TEST:
      return
    uniuni_reaction = self.simple.reactions[UNIUNI0]
    self.mesgraph.processUniUniReaction(uniuni_reaction)
    self.assertTrue(type(self.mesgraph.printSOMPath(DFG, E1))==str)
    self.assertFalse(self.mesgraph.printSOMPath(GLY, MEL))

  def testCheckTypeOneError(self):
    if IGNORE_TEST:
      return
    uniuni_reaction1 = self.simple.reactions[UNIUNI1]
    uniuni_reaction2 = self.simple.reactions[UNIUNI2]
    inequality_reaction1 = self.simple.reactions[INEQUAL1]
    inequality_reaction2 = self.simple.reactions[INEQUAL2]
    self.mesgraph.processUniUniReaction(uniuni_reaction1)
    self.mesgraph.processUniUniReaction(uniuni_reaction2)
    aa = self.simple.getMolecule(AA)
    cn = self.simple.getMolecule(CN)
    mg = self.simple.getMolecule(MG)
    self.assertTrue(self.mesgraph.checkTypeOneError((aa, cn), inequality_reaction1))
    self.assertFalse(self.mesgraph.checkTypeOneError((mg, aa), inequality_reaction2))
    self.assertTrue(len(self.mesgraph.type_one_errors)>0)
    self.assertFalse(len(self.mesgraph.type_two_errors)>0)

  def testCheckTypeTwoError(self):
    if IGNORE_TEST:
      return
    mesgraph2 = MESGraph(self.simple2)
    mesgraph2.processUniUniReaction(self.simple2.getReaction(REACTION1))
    mesgraph2.processMultiUniReaction(self.simple2.getReaction(REACTION2))
    mesgraph2.processMultiUniReaction(self.simple2.getReaction(REACTION3))
    self.assertTrue(mesgraph2.checkTypeTwoError())
    self.assertFalse(len(mesgraph2.type_one_errors)>0)
    self.assertTrue(len(mesgraph2.type_two_errors)>0)

  def testAnalyze(self):
    if IGNORE_TEST:
      return
    mesgraph1 = MESGraph(self.simple)
    mesgraph1.analyze(self.simple.reactions)
    self.assertEqual(len(mesgraph1.nodes), FINAL_NODES)
    self.assertEqual(len(mesgraph1.edges), FINAL_EDGES)
    self.assertTrue(len(mesgraph1.type_one_errors)>0)
    self.assertFalse(len(mesgraph1.type_two_errors)>0)
    #
    mesgraph2 = MESGraph(self.simple2)
    mesgraph2.analyze(self.simple2.reactions)
    self.assertTrue(len(mesgraph2.type_one_errors)>0)
    self.assertTrue(len(mesgraph2.type_two_errors)>0)
    #
    mesgraph3 = MESGraph(self.simple3)
    mesgraph3.analyze(self.simple3.reactions) 
    self.assertTrue(len(mesgraph3.type_three_errors)>0) 
    #
    mesgraph4 = MESGraph(self.simple4)
    mesgraph4.analyze(self.simple4.reactions) 
    self.assertTrue(len(mesgraph4.type_four_errors)>0)  
    #
    # mesgraph5 = MESGraph(self.simple5)
    # mesgraph5.analyze(self.simple5.reactions) 
    # dih = mesgraph5.getNode(self.simple5.getMolecule(DIH))
    # din = mesgraph5.getNode(self.simple5.getMolecule(DIN))
    # self.assertTrue(mesgraph5.has_edge(dih, din))  
    # self.assertTrue(mesgraph5.has_edge(din, dih)) 
    # self.assertEqual(len(mesgraph5.type_five_errors), 1)
    #   
    mesgraph6 = MESGraph(self.simple6)
    mesgraph6.analyze(self.simple6.reactions) 
    total_errors = len(mesgraph6.type_one_errors) + \
                   len(mesgraph6.type_two_errors) + \
                   len(mesgraph6.type_three_errors) + \
                   len(mesgraph6.type_four_errors) +\
                   len(mesgraph6.type_five_errors) 
    self.assertTrue(total_errors==0)

if __name__ == '__main__':
  unittest.main()




"""
Tests for Set Of Moletules (SOM)
"""
from SBMLLint.common import constants as cn
from SBMLLint.common.molecule import Molecule
from SBMLLint.common.reaction import Reaction
from SBMLLint.common.simple_sbml import SimpleSBML
from SBMLLint.games.som import SOM
from SBMLLint.common import simple_sbml

import numpy as np
import os
import re
import unittest


IGNORE_TEST = False
NUM_MERGED_SOMS = 2
NAMEFILTER = "[0-9a-zA-Z]+"


#############################
# Tests
#############################
class TestSOM(unittest.TestCase):

  def setUp(self):
    self.simple = SimpleSBML()
    self.simple.initialize(cn.TEST_FILE3)
    self.molecules = self.simple.molecules
    self.soms = []
    for mole in self.molecules:
      self.soms.append(SOM({mole}))
  
  def testConstructor(self):
    if IGNORE_TEST:
      return
    self.assertEqual(len(self.soms), len(self.molecules))
    self.assertEqual(self.soms[0].molecules, {self.molecules[0]})
    
  def testMakeId(self):
    if IGNORE_TEST:
      return
    som = self.soms[0]
    self.assertTrue(som.molecules.intersection(self.molecules))
    molecule = list(re.findall(NAMEFILTER, som.identifier))[0]
    self.assertEqual(list(som.molecules)[0], self.simple.getMolecule(molecule))
  
  def testMerge(self):
    if IGNORE_TEST:
      return
    som1 = self.soms[0]
    som2 = self.soms[1]
    molecule1 = list(som1.molecules)[0]
    molecule2 = list(som2.molecules)[0]
    new_som = som1.merge(som2)
    self.assertEqual(len(new_som.molecules), NUM_MERGED_SOMS)
    self.assertTrue(molecule1 in new_som.molecules)
    self.assertTrue(molecule2 in new_som.molecules)

if __name__ == '__main__':
  unittest.main()



""" Helpers for Tests. """

from SBMLLint.common import constants as cn

import os.path
import pandas as pd

DEBUG = False


def isValidDataFrame(df, expected_columns, min_rows=1,
    nan_columns=None, key=None,
    valid_dict=None):
  """
  Simple test of a data frame.
   a) Contains the expected columns
   b) Has a minimum number of rows
   c) Does not contain np.nan values
  :param pd.DataFrame df: DataFrame to validate
  :param list-of-str expected_columns:
  :param int min_rows:
  :param list-of-str nan_columns: columns where there may be nan
      values
  :param str or list-of-str key: Columns that is a key
  :param dict valid_dict: key=column name, value=function of value
  :return bool: True if passes tests
  """
  bads = [x for x in expected_columns if not x in df.columns.tolist()]
  if len(bads) > 0:
    if DEBUG:
      import pdb; pdb.set_trace()
    return False
  if len(df) < min_rows:
    if DEBUG:
      import pdb; pdb.set_trace()
    return False
  if (key is not None) and len(key) > 0:
    df_key = pd.DataFrame(df[key])
    if len(key) == 1:
      keys = df[key]
      if cn.NONE in keys:
        keys.remove(cn.NONE)
        df_key[key] = keys
    df_key = df_key.drop_duplicates()
    if len(df_key) != len(df.drop_duplicates()):
      if DEBUG:
        import pdb; pdb.set_trace()
      return False
  if valid_dict is not None:
    for col, func in valid_dict.items():
      trues = [func(x) for x in df[col]]
      if not all(trues):
        import pdb; pdb.set_trace()
        if DEBUG:
          import pdb; pdb.set_trace()
        return False
  return True



#!/usr/bin/env python
"""
Creates the moiety_structure section of the SBMLLint
configuration file. The user declares all moieties.
The moiety structure for a molecule is the number of
occurrences of each moiety string in the molecule.
Substrings are handled by looking first for the longest
moiety names.
"""

from SBMLLint.common.simple_sbml import SimpleSBML
from SBMLLint.common.moiety import MoietyStoichiometry
from SBMLLint.common import msgs

import argparse
import sys
import yaml


MOIETY_COUNT_SEPARATOR = ", "


def getMoieties(moiety_fid):
  """
  Creates the YAML configuration file.
  :param IOWrappter moiety_fid: 
  :return list-str: list of moiety names
  :raises ValueError: if one moiety is a substring of another
  """
  return yaml.safe_load(moiety_fid)

# TODO: Extend to do recursive checks of moiety names for
# their substructures
def findMoietyStoichiometries(molecule_name, moiety_names,
    is_check_error=True):
  """
  Finds the moieties and their stoichiometries for each molecule.
  Handles substrings by looking for largest strings first.
  :param str molecule:
  :param list-str moiety_names:
  :return list-MoietyStoichiometry:
  """
  result = []
  sorted_moiety_names = sorted(
      moiety_names, key=lambda n: len(str(n)), reverse=True)
  current_name = molecule_name
  for moiety_name in sorted_moiety_names:
    moiety_name = str(moiety_name)
    count = current_name.count(moiety_name)
    if count > 0:
      current_name = current_name.replace(moiety_name, "")
      result.append(MoietyStoichiometry(moiety_name, count,
          MOIETY_COUNT_SEPARATOR))
  current_name = current_name.replace("_", "")
  if is_check_error and len(current_name) > 0:
    msg = "Moieties specified do not completely express %s" % molecule_name
    msgs.error(msg)
  return result

# TODO: Look for moieties from largest string to smallest,
# eliminating those found.
def main(xml_fid, moiety_fid, otfile_fid):
  """
  Creates the YAML configuration file.
  :param IOWrapper xml_fid: xml file:
  :param IOWrapper moiety_fid: 
  :param IOWrapper otfile_fid: file to write
  """
  # Acquire and validate moieties
  moiety_names = getMoieties(moiety_fid)
  # Create the configuration file
  config_dct = {} 
  simple = SimpleSBML()
  simple.initialize(xml_fid)
  for molecule in simple.molecules:
    moiety_stoichiometries = findMoietyStoichiometries(
       molecule.name, moiety_names)
    config_dct[molecule.name] =  \
        [str(ms) for ms in moiety_stoichiometries]
  dct = {"moiety_structure": config_dct}
  yaml.dump(dct, otfile_fid)
  for fid in [xml_fid, moiety_fid, otfile_fid]:
    fid.close()

    
if __name__ == '__main__':
  def open_write(path):
    return open(path, "w")
  #
  parser = argparse.ArgumentParser(
      description='Creates configuration file for BioModels 147.')
  parser.add_argument('xml_file', 
      type=open, 
      help='SBML file from which species names are obtained.')
  parser.add_argument('moiety_file', 
      type=open, 
      help='YAML file the specifies moieties for the XML model.')
  parser.add_argument('--output', type=open_write,
      help="Generated configuration file. Defaults to config.yml.")
  args = parser.parse_args()
  if args.output is None:
    otfile_fid = sys.stdout
  else:
    otfile_fid = args.output
  main(args.xml_file, args.moiety_file, otfile_fid)



from SBMLLint.common import constants as cn
from SBMLLint.tools import make_moiety_structure


import numpy as np
import os
import sys
import unittest
import yaml


IGNORE_TEST = False
TEST_MOIETY_FILE = os.path.join(cn.TEST_DIR,
    "test_moieties.yml")
TEST_MOIETY_REDUN_FILE = os.path.join(cn.TEST_DIR,
    "test_moieties_redun.yml")
# moiety_structure section of config file
TEST_CONFIG_FILE = os.path.join(cn.TEST_DIR,
     "test_cfg_file.yml")
TEST_SBML_FILE = os.path.join(cn.TEST_DIR,
    "test_BIOMD0000000147_url.xml")
MOIETIES = ["IkBa", "IkBb", "IkBe", "IKK", "NFkB",
    "nucleus", "cytoplasm", "mRNA"]
MOIETIES_REDUN = ["A", "AA"]


#############################
# Tests
#############################
class TestFunctions(unittest.TestCase):

  def _cleanUp(self):
    for ffile in [TEST_CONFIG_FILE, TEST_MOIETY_FILE,
        TEST_MOIETY_REDUN_FILE]:
      if os.path.isfile(ffile):
        os.remove(ffile)

  def setUp(self):
    self._cleanUp()
    with open(TEST_MOIETY_FILE, "w") as fd:
      yaml.dump(MOIETIES, fd)
    self.xml_fid = open(TEST_SBML_FILE)
    self.moiety_fid = open(TEST_MOIETY_FILE)

  def tearDown(self):
    self._cleanUp()
    self.xml_fid.close()
    self.moiety_fid.close()

  def testGetMoieties(self):
    if IGNORE_TEST:
      return
    names = make_moiety_structure.getMoieties(self.moiety_fid)
    diff = set(names).symmetric_difference(MOIETIES)
    self.assertEqual(len(diff), 0)

  def testFindMoietyStoichiometries(self):
    if IGNORE_TEST:
      return
    molecule_name = "AAB"
    moiety_names = ["A", "AA", "B"]
    result = make_moiety_structure.findMoietyStoichiometries(
        molecule_name, moiety_names)
    self.assertEqual(result[0].moiety.name, "AA")
    self.assertEqual(result[0].stoichiometry, 1)
    self.assertEqual(len(result), 2)
    #
    molecule_name = "AAAB"
    result = make_moiety_structure.findMoietyStoichiometries(
        molecule_name, moiety_names)
    self.assertEqual(len(result), 3)

  def testMain(self):
    if IGNORE_TEST:
      return
    def isSubstr(stg, stgs):
      return any([stg in s for s in stgs])
    #
    config_fid = open(TEST_CONFIG_FILE, "w")
    make_moiety_structure.main(self.xml_fid,
        self.moiety_fid, config_fid)
    self.assertTrue(os.path.isfile(TEST_CONFIG_FILE))
    with open(TEST_CONFIG_FILE, "r") as fd:
      dct = yaml.safe_load(fd)
    moiety_dct = dct[cn.CFG_MOIETY_STRUCTURE]
    missing = [m for m in MOIETIES 
        if not isSubstr(m, moiety_dct.keys())]
    self.assertEqual(len(missing), 0)
    

if __name__ == '__main__':
  unittest.main()



"""
Molecule in a chemical reaction and MoleculeStoichiometry
(a molecule with its stoichiometry).

A Molecule is structured as one or more MoietyStoichiometry
with a separator. MOIETY_DOUBLE_SEPARATOR is used if at least
one MoietyStoichiometry has a repetition count; otherwise,
either MOIETY_DOUBLE_SEPARATOR or MOIETY_SEPRATOR can be used.

MOLECULE           MOIETY, STOICHIOMETRY
 A                  (A, 1)
 A_P_P_P            (A, 1), (P, 3)
 A__P__P__P         (A, 1), (P, 3)
 A__P_3             (A, 1), (P, 3)
"""

from SBMLLint.common import constants as cn
from SBMLLint.common import config
from SBMLLint.common.moiety import Moiety, MoietyStoichiometry
from SBMLLint.common import util

import pandas as pd
import numpy as np


class Molecule(object):

  def __init__(self, name):
    """
    :param str name:
    """
    self.name = name
    self._moiety_stoichiometrys = None

  @property
  def moiety_stoichiometrys(self):
    done = False
    if self._moiety_stoichiometrys is None:
      config_dct = config.getConfiguration()
      if cn.CFG_MOIETY_STRUCTURE in config_dct:
        dct = config_dct[cn.CFG_MOIETY_STRUCTURE]
        if self.name in dct.keys():
          self._moiety_stoichiometrys =  \
              MoietyStoichiometry.makeFromDct(dct[self.name])
          done = True
    else:
      done = True
    if not done:
      new_name = self._reformat()
      stgs = new_name.split(cn.MOIETY_DOUBLE_SEPARATOR)
      result = [MoietyStoichiometry.make(ms) for ms in stgs]
      result.sort()
      self._moiety_stoichiometrys = result
    return self._moiety_stoichiometrys

  def __repr__(self):
    return self.name

  def __lt__(self, other):
    return self.name < other.name

  def isEqual(self, other):
    return self.name == other.name

  def getMoietys(self):
    """
    Extracts the unique moieties in the molecule.
    :return list-Moiety: Unique Moiety in molecule
    """
    names = list(set([m_s.moiety.name for m_s in self.moiety_stoichiometrys]))
    names.sort()
    return [Moiety(n) for n in names]

  def _reformat(self):
    """
    Reformats the molecule name to use MOIETY_DOUBLE_SEPARATOR.
    :return str:
    """
    new_name = None
    pos = self.name.find(cn.MOIETY_DOUBLE_SEPARATOR)
    if pos > 0:
      new_name = self.name
    else:
      # Check to see if there is a single moiety
      parts = self.name.split(cn.MOIETY_SEPARATOR)
      if len(parts) == 2:
        if util.isInt(parts[1]):
          new_name = self.name
      if new_name is None:
        new_name = self.name.replace(cn.MOIETY_SEPARATOR,
            cn.MOIETY_DOUBLE_SEPARATOR)
    return new_name

  def append(self, element):
    """
    Appends to the end of a molecule.
    :param MoietyStoichiometry or Molecule element:
    :return Molecule:
    """
    new_name = "%s%s%s" % (
        self.name, cn.MOIETY_DOUBLE_SEPARATOR, 
        element.name)
    return Molecule(new_name)

  def hasMoiety(self, moiety):
    """
    :param Moiety moiety:
    :return bool: True if moiety name in molecule
    """
    moietys = self.getMoietys()
    return any([moiety.isEqual(m) for m in moietys])


class MoleculeStoichiometry(object):

  def __init__(self, molecule, stoichiometry):
    if not isinstance(molecule, Molecule):
      raise ValueError("First argument must be a Molecule.")
    if not util.isFloat(stoichiometry):
      raise ValueError("Second argument must be a float.")
    self.molecule = molecule
    self.stoichiometry = stoichiometry

  def __repr__(self):
    return "%s * % 2.2f" % (str(self.molecule), self.stoichiometry)

  def __lt__(self, other):
    return str(self) < str(other)

  def countMoietys(self):
    """
    Counts the occurrence of moietys.
    :return pd.DataFrame: index is moiety, value is count
    """
    moiety_stoichs = self.molecule.moiety_stoichiometrys
    moietys = list([str(m.moiety) for m in moiety_stoichs])
    stoichs = list([m.stoichiometry for m in moiety_stoichs])
    df = pd.DataFrame({cn.MOIETY: moietys, cn.VALUE: stoichs})
    df_result = pd.DataFrame(df.groupby(cn.MOIETY).sum())
    df_result = df_result.rename(
        columns={df_result.columns.tolist()[0]: cn.VALUE})
    df_result[cn.VALUE] = self.stoichiometry*df_result[cn.VALUE]
    return df_result

  @classmethod
  def countMoietysInCollection(cls, molecule_stoichiometrys):
    """
    Counts the occurrence of moietys.
    :param list-MoleculeStoichiometry  molecule_stoichiometrys:
    :return pd.DataFrame: cn.VALUE, indexed by moiety.name
    """
    dfs = []
    for molecule_stoichiometry in molecule_stoichiometrys:
      dfs.append(molecule_stoichiometry.countMoietys())
    df = pd.concat(dfs)
    df = df.reset_index()
    col = df.columns[0]
    df_result = pd.DataFrame(df.groupby(col).sum())
    return df_result

  @classmethod
  def getMolecules(cls, molecule_stoichiometrys):
     """
     Obtainins the molecules from a collection of
     MoleculeStoichiometry.
     :param list-MoleculeStoichiometry molecule_stoichiometrys;
     :return list-Molecule: unique molecules
     """
     molecules = [m_s.molecule for m_s in molecule_stoichiometrys] 
     return util.uniqueify(molecules)



from SBMLLint.common import constants as cn
from SBMLLint.tools import lp_analysis
from SBMLLint.common import simple_sbml


import numpy as np
import os
import sys
import unittest
import yaml


IGNORE_TEST = False
TEST_SBML_INCONSISTENT_PTH = os.path.join(cn.TEST_DIR,
    "test_BIOMD0000000147_url.xml")
TEST_SBML_CONSISTENT_PTH = os.path.join(cn.TEST_DIR,
    "test_BIOMD0000000145_url.xml")
TEST_SBML_CONSISTENT_PTH = os.path.join(cn.TEST_DIR,
    "test_BIOMD0000000010_url.xml")

#############################
# Tests
#############################
class TestFunctions(unittest.TestCase):

  def testLPAnalysis(self):
    if IGNORE_TEST:
      return
    def test(path, expected):
      with open(path, "r") as fd:
        self.assertEqual(
            lp_analysis.LPAnalysis(fd, is_report=True), expected)
    #
    test(TEST_SBML_INCONSISTENT_PTH, False)
    test(TEST_SBML_CONSISTENT_PTH, True)
    

if __name__ == '__main__':
  unittest.main()



"""Constants for Project."""
import collections
import os

PROJECT_NAME = "SBMLLint"

MOIETY_ANALYSIS = "moiety_analysis"
GAMES = "games"

############### COLUMN NAMES ##############
FILENAME = "filename"
IS_STRUCTURED = "is_structured"
NUM_BOUNDARY_REACTIONS = "num_boundary_reactions"
TOTAL_REACTIONS = "total_reactions"
NUM_IMBALANCED_REACTIONS = "num_imbalanced_reactions"
NUM_BALANCED_REACTIONS = "num_balanced_reactions"
FRAC_BALANCED_REACTIONS = "frac_balanced_reactions"
FRAC_BOUNDARY_REACTIONS = "frac_Boundary_reactions"

################ DIRECTORIES #################
PROJECT_DIR = os.path.dirname(os.path.abspath(__file__))
for _ in range(2):
  PROJECT_DIR = os.path.dirname(PROJECT_DIR)
CODE_DIR = os.path.join(PROJECT_DIR, PROJECT_NAME)
TEST_DIR = os.path.join(PROJECT_DIR, "tests")

################ DATA DIRECTORIES #################
BIOMODELS_DIR = os.path.join(PROJECT_DIR, "data/biomodels")
BIOMODELS_ZIP_FILENAME = "biomodels.zip"
BIGG_DIR = os.path.join(PROJECT_DIR, "data/bigg")
ANALYSIS_MOIETY_ANALYSIS_DIR = os.path.join(PROJECT_DIR, "analysis")
ANALYSIS_MOIETY_ANALYSIS_DIR = os.path.join(
    ANALYSIS_MOIETY_ANALYSIS_DIR, "moiety_analysis")

############### TEST FILES #####################
TEST_FILE = os.path.join(TEST_DIR, "test_file.xml")
TEST_FILE2 = os.path.join(TEST_DIR, "test_file2.xml")
# test_file3: originally curated_017
TEST_FILE3 = os.path.join(TEST_DIR, "test_file3.xml")
TEST_FILE5 = os.path.join(TEST_DIR, "test_file5.antimony")
TEST_FILE4 = os.path.join(TEST_DIR, "test_file4.xml")
# test_file6: originally curated_050
TEST_FILE6 = os.path.join(TEST_DIR, "test_file6.xml")
# test_file7: originally curated_018
TEST_FILE7 = os.path.join(TEST_DIR, "test_file7.xml")
# test_file8: originally curated_004
TEST_FILE8 = os.path.join(TEST_DIR, "test_file8.xml")
# test_file8: originally curated_008
TEST_FILE9 = os.path.join(TEST_DIR, "test_file9.xml")
# test_file10: originally "BIOMD0000000231_url.xml"
TEST_FILE10 = os.path.join(TEST_DIR, "test_file10.xml")
# test_file10: originally "BIOMD0000000253_url.xml"
TEST_FILE11 = os.path.join(TEST_DIR, "test_file11.xml")
# test_file12: originally "BIOMD0000000281_url.xml"
TEST_FILE12 = os.path.join(TEST_DIR, "test_file12.xml")
# test_file13: originally "BIOMD0000000035_url.xml"
TEST_FILE13 = os.path.join(TEST_DIR, "test_file13.xml")
# test_file_games_pp1: originally "BIOMD0000000383_url.xml"
TEST_FILE_GAMES_PP1 = os.path.join(TEST_DIR, "test_file_games_pp1.xml")
# test_file_games_pp2: originally "BIOMD0000000018_url.xml"
TEST_FILE_GAMES_PP2 = os.path.join(TEST_DIR, "test_file_games_pp2.xml")
# test_file_games_report1: originally "BIOMD0000000248_url.xml"
TEST_FILE_GAMESREPORT1 = os.path.join(TEST_DIR, "test_file_games_report1.xml")
# test_file_games_report2: originally "BIOMD0000000007_url.xml"
TEST_FILE_GAMESREPORT2 = os.path.join(TEST_DIR, "test_file_games_report2.xml")
# test_file_games_report3: originally "BIOMD0000000167_url.xml"
TEST_FILE_GAMESREPORT3 = os.path.join(TEST_DIR, "test_file_games_report3.xml")
NUM_REACTIONS = 111
NUM_PARAMETERS = 27
MAX_REACTANTS = 10
NUM_SPECIES = 32

############## FUNCTIONAL CONSTANTS ##############
# DataFrame Columns
VALUE = "value"
MOIETY = "moiety"

#
ARC_ARROW = "->"
EQUAL = "="
LABEL_SEPARATOR = ":"
LESSTHAN = "<"
KINETICS_SEPARATOR = ";"
MOIETY_SEPARATOR = "_"
MOIETY_DOUBLE_SEPARATOR = MOIETY_SEPARATOR + MOIETY_SEPARATOR

# Reaction categories
REACTION_1_1 = "reaction_1_1"
REACTION_n_1 = "reaction_n_1"
REACTION_1_n = "reaction_1_n"
REACTION_n_n = "reaction_n_n"
REACTION_BOUNDARY = "reaction_boundary"
# for reduced reactions
REACTION_REDUNDANT = "reaction_redundant"
REACTION_ERROR = "reaction_error"
ReactionCategory = collections.namedtuple('ReactionCategory',
    'category predicate')
ReactionComponents = collections.namedtuple('ReactionComponents',
    'label reactants products')

# EmptySet in a reaction (ex. curated model 006)
EMPTYSET = "EmptySet"

# Reaction attributes for each arc in MESGraph
REACTION = "reaction"

# The selected category is the first one that first has
# a satisfied predicate
# Arguments:
# x: number of reactants, y: number of products
# z: sum of reactants stoichiometry
# w: sum of products stoichiometry
REACTION_CATEGORIES = [
    ReactionCategory(category=REACTION_1_1,
        predicate=lambda x,y,z,w: (x==1) and (y==1) and (z==w)),
    ReactionCategory(category=REACTION_1_n,
        predicate=lambda x,y,z,w: ((x==1) and (y>1) and (z==1.00)) \
                               or ((x==1) and (y==1) and (z<w))),
    ReactionCategory(category=REACTION_n_1,
        predicate=lambda x,y,z,w: ((x>1) and (y==1) and (w==1.00)) \
                               or ((x==1) and (y==1) and (z>w))),
    ReactionCategory(category=REACTION_n_n,
        predicate=lambda x,y,z,w: ((x>1) and (y>1)) 
                               or ((x==1) and (y>1) and (z!=1.00)) \
                               or ((x>1) and (y==1) and (w!=1.00))),
    ReactionCategory(category=REACTION_BOUNDARY,
        predicate=lambda x,y,z,w: (x==0) or (y==0) or (z==0) or (w==0)),
    ]
# Similar to REACTION_CATEGORIES but have different arguments r, p 
# and two additional categories - reaction_redundant and reaction_error
# r: list of reactants stoichiometry
# p: list of products stoichiometry
REACTION_SUMMARY_CATEGORIES = [
    ReactionCategory(category=REACTION_REDUNDANT,
        predicate=lambda x,y,r,p: (x==0) and (y==0)),
    ReactionCategory(category=REACTION_ERROR,
        predicate=lambda x,y,r,p: ((x==0) and (y!=0)) \
                                  or ((x!=0) and (y==0))),
    ReactionCategory(category=REACTION_1_1,
        predicate=lambda x,y,r,p: (x==1) and (y==1) and (sum(r)==sum(p))),
    
    ReactionCategory(category=REACTION_1_n,
        predicate=lambda x,y,r,p: (x==1) and (sum([r[0]<=e for e in p])==len(p))),
                     
    ReactionCategory(category=REACTION_n_1,
        predicate=lambda x,y,r,p: (y==1) and (sum([p[0]<=e for e in r])==len(r))),
    ]


# summarized version of a reaction
ReactionSummary = collections.namedtuple('ReactionSummary', 
                  'label reactants products category')


# building block for each MESGraph path
PathComponents = collections.namedtuple('PathComponents',
                                        'node1 node2 reactions')

# used for creating a report
NULL_STR = ""

# Top level keys in configuration file
CFG_IGNORED_MOIETIES = "ignored_moieties"
CFG_IGNORED_MOLECULES = "ignored_molecules"
CFG_MOIETY_STRUCTURE = "moiety_structure"
CFG_PROCESS_BOUNDARY_REACTIONS = "process_boundary_reactions"
CFG_GAMES_THRESHOLD = "games_threshold_num_reactions"
CFG_SECTIONS = [
    CFG_IGNORED_MOLECULES,
    CFG_IGNORED_MOIETIES,
    CFG_PROCESS_BOUNDARY_REACTIONS,
    CFG_MOIETY_STRUCTURE,
    CFG_GAMES_THRESHOLD,
    ]

# Default values for configuration file
CFG_DEFAULTS = {}
CFG_DEFAULTS[CFG_IGNORED_MOIETIES] = ['DUMMYMOIETY']
CFG_DEFAULTS[CFG_IGNORED_MOLECULES] = ['DUMMYMOLECULE']
CFG_DEFAULTS[CFG_PROCESS_BOUNDARY_REACTIONS] = False
CFG_DEFAULTS[CFG_GAMES_THRESHOLD] = 20
CFG_DEFAULT_PATH = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
CFG_DEFAULT_PATH = os.path.join(CFG_DEFAULT_PATH, ".sbmllint_cfg.yml")



from SBMLLint.common import constants as cn
from SBMLLint.common.tellurium_sandbox import TelluriumSandbox
from SBMLLint.common import tellurium_sandbox as ts
from SBMLLint.common import exceptions


import numpy as np
import os
import unittest


IGNORE_TEST = False
INPUT = """
Hello World!
Hello World2!
"""
NUM_S1 = 2
NUM_S2 = 3
IGNORE_TEST = False
ANTIMONY_STG = '''
%dS1 -> %dS2; 1
S1 = 0
S2 = 0
''' % (NUM_S1, NUM_S2)


#############################
# Tests
#############################
class TestTelluriumSandbox(unittest.TestCase):

  def testRun(self):
    try:
      sandbox = TelluriumSandbox()
      sandbox.run("echo", INPUT)
      self.assertEqual(sandbox.return_code, 0)
      self.assertEqual(sandbox.output, INPUT)
    except exceptions.MissingTelluriumError:
      pass

  def testInstalledPackages(self):
    pkgs = ts.getInstalledPackages()
    self.assertTrue("scipy" in pkgs)
  
  def testGetSBMLFromAntimony(self):
    try:
      sandbox = TelluriumSandbox()
      sandbox.run("getSBMLFromAntimony", ANTIMONY_STG)
      self.assertEqual(sandbox.return_code, 0)
    except exceptions.MissingTelluriumError:
      pass
  
  def testConstructorMissingPakcage(self):
    with self.assertRaises(exceptions.MissingTelluriumError):
      sandbox = TelluriumSandbox(dependencies=["dummy"])


if __name__ == '__main__':
  unittest.main()



"""
Tests for MoietyComparator
"""
from SBMLLint.common import constants as cn
from SBMLLint.common import exceptions
from SBMLLint.common import config
from SBMLLint.common.molecule import Molecule, MoleculeStoichiometry
from SBMLLint.common.simple_sbml import SimpleSBML
from SBMLLint.common import util
from SBMLLint.moiety_analysis.moiety_comparator  \
     import MoietyComparator

import copy
import itertools
import numpy as np
import os
import unittest


IGNORE_TEST = False
MOIETY_NAME1 = "first"
MOIETY_NAME2 = "second"
MOIETY_NAME3 = "third"
NAMES = [MOIETY_NAME1, MOIETY_NAME2, MOIETY_NAME3]
MOLECULE_NAME = "%s%s%s" % (MOIETY_NAME1, cn.MOIETY_SEPARATOR, 
    MOIETY_NAME2)
iterator = itertools.product([0,1], [0, 1], [0, 1])
MOLECULE_NAMES = []  # A set of names from moiety combinations
for item in iterator:
  name = ""
  for idx,ele in enumerate(item):
    if ele == 1:
      if len(name) == 0:
        name = NAMES[idx]
      else:
        name = "%s%s%s" % (name, cn.MOIETY_SEPARATOR, NAMES[idx])
  if len(name) > 0:
    MOLECULE_NAMES.append(name)
MOLECULE_NAMES1 = MOLECULE_NAMES[:3]
MOLECULE_NAMES2 = MOLECULE_NAMES[3:]
TEST_FILE4 = "test_file4.xml"
TEST_FILE3 = "test_file3.antimony"
# Antimony as source
PATH= os.path.join(cn.TEST_DIR, TEST_FILE3)
with open(PATH, 'r') as fd:
  lines = fd.readlines()
NUM1 = 2
NUM2 = 3


######################################
# Auxiliary Functions
######################################
def analyze(simple):
  return MoietyComparator.analyzeReactions(simple)
  

#############################
# Tests
#############################
class TestMoietyComparator(unittest.TestCase):

  def setUp(self):
    self.molecules1 = [MoleculeStoichiometry(Molecule(n), NUM1)
        for n in MOLECULE_NAMES1]
    self.molecules2 = [MoleculeStoichiometry(Molecule(n), NUM2)
        for n in MOLECULE_NAMES2]
    self.comparator = MoietyComparator(self.molecules1,
        self.molecules2)
    self.config_dict = copy.deepcopy(config._config_dict)

  def tearDown(self):
    config._config_dict = self.config_dict

  def testConstructor(self):
    if IGNORE_TEST:
      return
    self.assertEqual(
        len(self.comparator.molecule_stoichiometry_collections), 2)

  def testIsSame(self):
    if IGNORE_TEST:
      return
    self.assertFalse(self.comparator.isSame())
    comparator = MoietyComparator(self.molecules1, self.molecules1)
    self.assertTrue(comparator.isSame())

  def testDifference1(self):
    if IGNORE_TEST:
      return
    comparator = MoietyComparator(self.molecules1, self.molecules2)
    df = comparator.difference()
    self.assertLess(df.loc[MOIETY_NAME1].tolist()[0], 0)

  def testDifference2(self):
    if IGNORE_TEST:
      return
    config._config_dict[cn.CFG_IGNORED_MOIETIES] = [MOIETY_NAME1]
    comparator = MoietyComparator(self.molecules1,
        self.molecules2)
    df = comparator.difference()
    self.assertFalse(MOIETY_NAME1 in df.index)

  def testDifference3(self):
    if IGNORE_TEST:
      return
    config._config_dict[cn.CFG_IGNORED_MOLECULES] = [MOIETY_NAME1]
    comparator = MoietyComparator(self.molecules1,
        self.molecules2)
    df = comparator.difference()
    expected_count = sum([NUM1 if MOIETY_NAME1 in m else 0 
        for m in MOLECULE_NAMES1 if len(MOIETY_NAME1) != len(m)])
    expected_count -= sum([NUM2 if MOIETY_NAME1 in m else 0 
        for m in MOLECULE_NAMES2 if len(MOIETY_NAME1) != len(m)])
    self.assertEqual(df.loc[MOIETY_NAME1, cn.VALUE], expected_count)

  def testDifference4(self):
    if IGNORE_TEST:
      return
    config._config_dict[cn.CFG_PROCESS_BOUNDARY_REACTIONS] = False
    molecules1 = [MoleculeStoichiometry(Molecule(n), 0)
        for n in MOLECULE_NAMES[:3]]
    comparator = MoietyComparator(molecules1, self.molecules2)
    df = comparator.difference()
    self.assertEqual(df[df.columns[0]].sum(), 0)

  def testDifference5(self):
    if IGNORE_TEST:
      return
    config._config_dict[cn.CFG_PROCESS_BOUNDARY_REACTIONS] = True
    molecules1 = [MoleculeStoichiometry(Molecule(n), 0)
        for n in MOLECULE_NAMES[:3]]
    comparator = MoietyComparator(molecules1, self.molecules2)
    df = comparator.difference()
    self.assertNotEqual(df[df.columns[0]].sum(), 0)
  
  def testReportDifference(self):  
    if IGNORE_TEST:
      return
    stg = self.comparator.reportDifference()
    self.assertGreater(len(stg), 0)
    comparator = MoietyComparator(self.molecules1,
        self.molecules1)
    stg = comparator.reportDifference()
    self.assertEqual(len(stg), 0)

  def testAnalyzeReactions(self):
    if IGNORE_TEST:
      return
    simple = SimpleSBML()
    try:
      SBML= util.getXMLFromAntimony(''.join(lines))
    except exceptions.MissingTelluriumError:
      return
    simple.initialize(SBML)
    result = analyze(simple)
    self.assertGreaterEqual(result.num_reactions, 0)
    self.assertGreaterEqual(result.num_imbalances, 0)
    self.assertTrue('2' in result.report)
    self.assertGreater(result.report.count('\n'),  5)
    

if __name__ == '__main__':
  unittest.main()



from SBMLLint.common import constants as cn
from SBMLLint.common.runner import Runner
from SBMLLint.tools import analyze_moiety_analysis


import numpy as np
import os
import unittest


IGNORE_TEST = False
TEST_FILE = "test_analyze_moiety_analysis.csv"
TEST_OUT_PATH = os.path.join(cn.TEST_DIR, TEST_FILE)


#############################
# Tests
#############################
class TestFunctions(unittest.TestCase):

  def tearDown(self):
    if os.path.isfile(TEST_OUT_PATH):
      os.remove(TEST_OUT_PATH)

  def testIsStructuredName(self):
    self.assertTrue(analyze_moiety_analysis.isStructuredName("a_b"))
    self.assertFalse(analyze_moiety_analysis.isStructuredName("a_1"))
    self.assertFalse(analyze_moiety_analysis.isStructuredName("Species_1"))
    self.assertFalse(analyze_moiety_analysis.isStructuredName("ATP"))

  def testCalcStats(self):
    analyze_moiety_analysis.calcStats(initial=1, final=5,
         out_path=TEST_OUT_PATH, report_progress=False)
    self.assertTrue(os.path.isfile(TEST_OUT_PATH))



if __name__ == '__main__':
  unittest.main()



"""A Tool that creates data to analyze the use of names
   that are structured to expose moieties.
"""

from SBMLLint.common import constants as cn
from SBMLLint.common.molecule import Molecule
from SBMLLint.common import simple_sbml
from SBMLLint.common import util
from SBMLLint.common.reaction import Reaction
from SBMLLint.tools import sbmllint
from SBMLLint.tools import print_reactions

import os
import numpy as np
import pandas as pd

# File paths
OUTPUT_FILE = "analyze_moiety_analysis.csv"
DIR = os.path.dirname(os.path.abspath(__file__))
OUTPUT_PATH = os.path.join(DIR, OUTPUT_FILE)
# Miscellaneous
EXCLUDE_PREFIX = ["node", "x", "species"]
EXCLUDE_SUFFIX = ["prime", "mrna", "prot"]

def isStructuredName(name):
  """
  Detects if the name is a candidate for a moiety structured name.
  The term "element" refers to substrings separated by "_".
  Not a structured name if any of the following are True:
    No "_" present
    Last element is a number
    First element is in EXCLUDE_PREFIX (case independent)
    Last element is in EXCLUDE_SUFFIX (case independent)
  """
  moietys = name.split(cn.MOIETY_SEPARATOR)
  if len(moietys) == 1:
    return False
  # See if last element has a numeric suffix
  try:
    _ = float(moietys[-1])
    return False
  except:
      pass
  if any([moietys[0].lower() in s for s in EXCLUDE_PREFIX]):
    return False
  if any([moietys[-1].lower() in s for s in EXCLUDE_SUFFIX]):
    return False
  return True

def calcStats(initial=0, final=50, out_path=OUTPUT_PATH, 
    report_interval=50, report_progress=True, min_frc=-1,
    data_dir=cn.BIOMODELS_DIR):
  """
  Calculates statistics for structured names.
  :param int initial: Index of first model to process
  :param int final: Index of final model to process
  :param str out_path: Path to the output CSV file
  :param int report_interval: Number of files processed before
      a report is written
  :param bool report_progress: report file being processed
  :param float min_frc: Filter to select only those models
      that have at least the specified fraction of reactions
      balanced according to moiety_analysis
  """
  def writeDF(dfs):
    df_count = pd.concat(dfs)
    df_count[cn.NUM_BALANCED_REACTIONS] =  \
        df_count[cn.TOTAL_REACTIONS]  \
        - df_count[cn.NUM_IMBALANCED_REACTIONS]
    denom =  (df_count[cn.TOTAL_REACTIONS] 
        - df_count[cn.NUM_BOUNDARY_REACTIONS])
    denom = [np.nan if np.isclose(v, 0) else v for v in denom]
    df_count[cn.FRAC_BALANCED_REACTIONS] =  \
        1.0*df_count[cn.NUM_BALANCED_REACTIONS] / denom
    df_count[cn.FRAC_BOUNDARY_REACTIONS] =  \
        1.0*df_count[cn.NUM_BOUNDARY_REACTIONS] / (
        df_count[cn.TOTAL_REACTIONS])
    if min_frc < 0:
      df = df_count
    else:
      df = df_count[df_count[cn.FRAC_BALANCED_REACTIONS] > min_frc]
    df = df.sort_values(cn.FRAC_BALANCED_REACTIONS)
    df.to_csv(out_path, index=False)
  #
  dfs = []
  sbmliter = simple_sbml.modelIterator(initial=initial, final=final,
      data_dir=data_dir)
  for item in sbmliter:
    if report_progress:
      print("*Processing file %s, number %d"
           % (item.filename, item.number))
    simple = simple_sbml.SimpleSBML()
    try:
      simple.initialize(item.model)
    except:
      print("  Error in model number %d." % item.number)
      continue
    row = {cn.FILENAME: [item.filename], 
           cn.IS_STRUCTURED: [False], 
           cn.NUM_BOUNDARY_REACTIONS: [0],
           cn.TOTAL_REACTIONS: [0],
           cn.NUM_IMBALANCED_REACTIONS: [0],
           }
    for reaction in simple.reactions:
      if (len(reaction.reactants) == 0) or (len(reaction.products) == 0):
          row[cn.NUM_BOUNDARY_REACTIONS] =  \
              [row[cn.NUM_BOUNDARY_REACTIONS][0] + 1]
      molecules = util.uniqueify([m.molecule 
          for m in set(reaction.reactants).union(reaction.products)])
      if any([isStructuredName(m.name) for m in molecules]):
          row[cn.IS_STRUCTURED] = [True]
    try:
      mcr = sbmllint.lint(model_reference=item.model, is_report=False)
      row[cn.TOTAL_REACTIONS] = [mcr.num_reactions if mcr.num_reactions > 0 else np.nan]
      row[cn.NUM_IMBALANCED_REACTIONS] = [mcr.num_imbalances]
    except:
      row[cn.TOTAL_REACTIONS] = [None]
      row[cn.NUM_IMBALANCED_REACTIONS] = [0]
    dfs.append(pd.DataFrame(row))
    if item.number % report_interval == 0:
      writeDF(dfs)
  writeDF(dfs)


if __name__ == '__main__':
  calcStats(initial=0, final=1000, report_interval=1,
      data_dir=cn.BIGG_DIR)



"""
Reporting class for GAMES Plus (GAMES_PP) algorithm
"""
from SBMLLint.common import constants as cn
from SBMLLint.common.molecule import Molecule, MoleculeStoichiometry
from SBMLLint.common.reaction import Reaction
from SBMLLint.common.simple_sbml import SimpleSBML
from SBMLLint.games.games_pp import SOMStoichiometry, SOMReaction, GAMES_PP, TOLERANCE
from SBMLLint.games.som import SOM
from SBMLLint.common import simple_sbml

import collections
import networkx as nx
import numpy as np
import pandas as pd

NULL_STR = ""
ReactionOperation = collections.namedtuple("ReactionOperation", 
    "reaction operation")
NUM_STAR = 70
PARAGRAPH_DIVIDER = "\n" + "-"*NUM_STAR + "\n"
REPORT_DIVIDER = "\n" + "*"*NUM_STAR + "\n"


class SimplifiedReaction(object):

  def __init__(self, reactants, products, label, mesgraph):
    """
    :param list-MoleculeStoichiometry reactants:
    :param list-MoleculeStoichiometry products:
    :param str label:
    """
    self.reactants = reactants
    self.products = products
    self.label = label
    self.identifier = self.makeIdentifier()
    self.mesgraph = mesgraph

  def __repr__(self):
    return self.identifier
    
  def makeIdentifier(self):
    """
    Provides a string representation of the reaction
    :param bool is_include_kinetics: include the kinetics formula
    :return str:
    """
    def makeStoichiometryString(molecule_stoichiometry):
      num = molecule_stoichiometry.stoichiometry
      if np.isclose(num, 1.0):
        return ''
      else:
        return "%2.2f " % num
    #
    def makeTermCollection(molecule_stoichiometries):
      """
      Formats a set of terms with stoichiometries.
      :param list-MoleculeStoichiometry:
      :return str:
      """
      term_collection = ''
      for m_s in molecule_stoichiometries:
        term = "%s%s" % (makeStoichiometryString(m_s), str(m_s.molecule))
        if len(term_collection) == 0:
          term_collection += term
        else:
          term_collection += " + " + term
      return term_collection
    #
    reactant_collection = makeTermCollection(self.reactants)
    product_collection = makeTermCollection(self.products)
    formula_str = ''
    reaction_str = "%s: %s -> %s" % (self.label,
        reactant_collection, product_collection)
    reaction_str = reaction_str + formula_str
    return reaction_str

  def reduceBySOMs(self):
    """
    Cancel reactants and products by SOMs
    SOMs is given by existing MESGraph.
    Return a new identifier. 
    :return str:
    """
    reactant_soms = {self.mesgraph.getNode(r.molecule) for r in self.reactants}
    product_soms = {self.mesgraph.getNode(p.molecule) for p in self.products}
    common_soms = list(reactant_soms.intersection(product_soms))
    if common_soms:
      for som in common_soms:
        reactants_in = collections.deque([ms for ms in
                                          self.reactants if
                                          self.mesgraph.getNode(ms.molecule)==som])
        reactants_out = [ms for ms in
                         self.reactants if
                         self.mesgraph.getNode(ms.molecule)!=som]
        products_in = collections.deque([ms for ms in
                                         self.products if
                                         self.mesgraph.getNode(ms.molecule)==som])
        products_out = [ms for ms in
                        self.products if
                        self.mesgraph.getNode(ms.molecule)!=som]
        #
        while reactants_in and products_in:
          reactant = reactants_in[0]
          product = products_in[0]
          if reactant.stoichiometry > product.stoichiometry:
            reactants_in[0] = MoleculeStoichiometry(reactant.molecule,
                                                    reactant.stoichiometry - product.stoichiometry)
            products_in.popleft()
          elif reactant.stoichiometry < product.stoichiometry:
            products_in[0] = MoleculeStoichiometry(product.molecule,
                                                   product.stoichiometry - reactant.stoichiometry)
            reactants_in.popleft()
          else:
            reactants_in.popleft()
            products_in.popleft()
        reactants = list(reactants_in) + reactants_out
        products = list(products_in) + products_out
        #  
        if (len(self.reactants) > len(reactants)) | \
          (len(self.products) > len(products)):
          self.reactants = reactants
          self.products = products
      self.identifier = self.makeIdentifier()
    return self.identifier


class GAMESReport(object):

  def __init__(self, mesgraph, explain_threshold=20, errors=None):
    self.mesgraph = mesgraph
    self.errors = errors
    self.report_type_one_errors = NULL_STR
    self.explain_threshold = explain_threshold

  def getMoleculeEqualityPath(self, som, molecule1, molecule2):
    """
    Create an undirected graph between
    two molecules within a SOM
    and find the shortest path
    :param SOM som:
    :param Molecule mole1:
    :param Molecule mole2:
    :return PathComponents: som_path
    """   
    # molecule1 = mole1.name
    # molecule2 = mole2.name
    # construct undirected graph
    subg = nx.Graph()
    # here, every reaction is 1-1 reaction
    for reaction in list(som.reactions):
      node1 = reaction.reactants[0].molecule.name
      node2 = reaction.products[0].molecule.name
      if subg.has_edge(node1, node2):
        reaction_label = subg.get_edge_data(node1, node2)[cn.REACTION]
        # if reaction.label is not included in the attribute, add its label
        if reaction.label not in set(reaction_label):
          reaction_label = reaction_label + [reaction.label]
      else:
        reaction_label = [reaction.label]    
      subg.add_edge(node1, node2, reaction=reaction_label)
    path = [short_p for short_p in nx.shortest_path(subg, 
                                                    source=molecule1, 
                                                    target=molecule2
                                                    )
           ]
    som_path = []
    for idx in range(len(path)-1):
      edge_reactions = subg.get_edge_data(path[idx], path[idx+1])[cn.REACTION]
      som_path.append(cn.PathComponents(node1=path[idx], 
                                        node2=path[idx+1],
                                        reactions=edge_reactions
                                        )
                     )
    return som_path

  def getMoleculeEqualityPathReport(self, molecule_name1, molecule_name2, reaction_count, explain_details):
    """
    Print out shortest path between two molecules within a 
    same SOM. Molecule names (str) are arguments. 
    :param str molecule_name1:
    :param str molecule_name2:
    :param int reaction_count:
    :param bool explain_details:
    :return bool/str: path_report
    :return int: reaction_count
    """
    path_report = NULL_STR
    som1 = self.mesgraph.getNode(self.mesgraph.simple.getMolecule(molecule_name1))
    som2 = self.mesgraph.getNode(self.mesgraph.simple.getMolecule(molecule_name2))
    if som1 != som2:
      return False
    else:
      # in case molecule_name1 == molecule_name2, for example A -> A + B
      if molecule_name1 == molecule_name2:
      	if explain_details:
          path_report = path_report + "Clearly, %s %s %s\n" % (
              molecule_name1, cn.EQUAL, molecule_name2)
      else:
        som_path = self.getMoleculeEqualityPath(som1, molecule_name1, molecule_name2)
        for pat in som_path:
          if explain_details:
            path_report = path_report + "\n%s %s %s by reaction(s):\n" % (pat.node1, cn.EQUAL, pat.node2)
          for r in pat.reactions:
            som_reaction = self.mesgraph.simple.getReaction(r)
            reaction_count += 1
            path_report = path_report + "%d. %s\n" % (reaction_count, som_reaction.makeIdentifier(is_include_kinetics=False))
      return reaction_count, path_report

  def getMoleculeInequalityPathReport(self, molecule_name1, molecule_name2, reaction_names, reaction_count, explain_details):
  	"""
  	Print the reactions that infer inequality between molecules. 
  	Molecule name are given as arguments.
  	Especially, the mass of molecule 1 is less than that of molecule 2
  	:param str molecule_name1:
  	:param str molecule_name2:
  	:param list-str reaction_names:
  	:param int reaction_count:
  	:param bool explain_details
  	:return str: path_report
  	:return int: reaction_count
  	"""
  	path_report = NULL_STR
  	if explain_details:
  	  path_report = path_report + "%s %s %s by reaction(s):\n" % (molecule_name1, cn.LESSTHAN, molecule_name2)
  	for reaction_name in reaction_names:
  	  reaction_count += 1
  	  reaction = self.mesgraph.simple.getReaction(reaction_name)
  	  path_report = path_report + "%d. %s\n" % (reaction_count, reaction.makeIdentifier(is_include_kinetics=False))
  	return reaction_count, path_report

  def reportTypeOneError(self, type_one_errors, explain_details=False):
    """
    Generate report for Type I Errors. 
    Type I error occurs when there is a reaction
    that needs to add an arc between molecules,
    while the molecules are already included
    in the same SOM.
    :param list-PathComponents type_one_errors:
    :param bool explain_details:
    :return str: type_one_report
    :return list-int: error_num
    """
    report = NULL_STR
    error_num = []
    if len(type_one_errors) == 0:
      return report, error_num
    for pc in type_one_errors:
      mole1 = pc.node1
      mole2 = pc.node2
      reactions = pc.reactions
      reaction_count = 0
      reaction_count, equality_report = self.getMoleculeEqualityPathReport(mole1, mole2, reaction_count, explain_details)
      report = report + equality_report
      if explain_details:
        report = report + "\nHowever, "
      reaction_count, inequality_report = self.getMoleculeInequalityPathReport(mole1, mole2, reactions, reaction_count, explain_details)
      report = report + inequality_report
      report = report + "\n%s\n" % (PARAGRAPH_DIVIDER)
      error_num.append(reaction_count)
    report = report + "\n%s\n" % (REPORT_DIVIDER)
    return report, error_num

  def reportTypeTwoError(self, type_two_errors, explain_details=False):
  	"""
  	Generate report for Type II Errors.
  	Type II Error occurs when there is
  	a cycle between SOMs, which
  	should not happen.
  	:param list-(list-SOMs) type_two_errors:
  	:param bool explain_details:
  	:return str: type_two_report
  	:return list-int: error_num
  	"""
  	report = NULL_STR
  	error_num = []
  	if len(type_two_errors) == 0:
  	  return report, error_num
  	for cycle in type_two_errors:
  	  all_soms = set()
  	  reaction_count = 0
  	  report = report + "We detected a mass imbalance from the following reactions:\n"
  	  if explain_details:
  	    report = report + "\nThese uni-uni reactions created mass-equivalence.\n"
  	  for som in cycle:
  	  	all_soms.add(som)
  	  	for reaction in list(som.reactions):  	  	  	
  	  	  reaction_count += 1
  	  	  report = report + "\n%d. %s" % (reaction_count, reaction.makeIdentifier(is_include_kinetics=False))
  	  if explain_details:
  	  	report = report + "\n%s\n" % (PARAGRAPH_DIVIDER)
  	  	report = report + "The following reactions create mass-inequality.\n"
  	  inequality_reactions = []
  	  for som_pair in zip(cycle, cycle[1:] + [cycle[0]]):
  	    inequality_reactions = inequality_reactions + self.mesgraph.get_edge_data(som_pair[0], som_pair[1])[cn.REACTION]
  	  for r in inequality_reactions:
  	    reaction = self.mesgraph.simple.getReaction(r)
  	    reaction_count += 1
  	    report = report + "\n%d. %s" % (reaction_count, reaction.makeIdentifier(is_include_kinetics=False))
  	  if explain_details:
  	    report = report + "\n%s\n" % (PARAGRAPH_DIVIDER)
  	  if explain_details:
  	  	reaction_count = reaction_count - len(inequality_reactions) 	
  	  	# explain the mass equivalent pseudo reactions and the resulting sets
  	  	report = report + "Based on the reactions above, we have mass-equivalent pseudo reactions.\n"
  	  	for r in inequality_reactions:
  	  	  reaction = self.mesgraph.simple.getReaction(r)
  	  	  som_reaction = self.mesgraph.convertReactionToSOMReaction(reaction)
  	  	  reaction_count += 1
  	  	  report = report + "\n(pseudo %d.) %s" % (reaction_count, som_reaction.identifier)
  	  	report = report +  "\n%s\n" % (PARAGRAPH_DIVIDER)
  	  	report = report + "However, the above pseudo reactions imply the following inequalities:\n\n"
  	  	for som in cycle:
  	  	  report = report + "%s %s " % (som.identifier, cn.LESSTHAN)
  	  	report = report + "%s\n" % (cycle[0].identifier)
  	  	report = report + "\nThis indicates a mass conflict between reactions."
  	  error_num.append(reaction_count)
  	  report = report + "\n%s%s\n" % (PARAGRAPH_DIVIDER, PARAGRAPH_DIVIDER)
  	return report, error_num

  def convertOperationSeriesToReactionOperations(self, operation):
    """
    Convert an operation series to 
    a list of reaction operations.
    An 'operation' is either a row or column 
    of an operation matrix, 
    where both column and row indices are reactions. 
    :param pandas.Series operation:
    :return list-ReactionOperation: operations
    """
    operations = []
    # 
    nonzero_idx = np.array([idx for idx, val in enumerate(operation) if val != 0.0])
    nonzero_op = operation[nonzero_idx]
    for idx in range(len(nonzero_op)):
      reaction_op = ReactionOperation(reaction=nonzero_op.index[idx],
      	                              operation=nonzero_op[idx]
      	                              )
      operations.append(reaction_op)
    return operations

  def getOperationMatrix(self):
  	"""
  	Return an operation matrix (pandas DataFrame)
  	on the transposed stoichiometry matrix
  	:return pandas.DataFrame: operation_df
  	"""
  	operation_df = None
  	if self.mesgraph.lower_inverse is None:
  	  pass
  	elif self.mesgraph.rref_operation is None:
  	  operation_df = self.mesgraph.lower_inverse
  	else:
  	  operation_df = self.mesgraph.rref_operation.dot(self.mesgraph.lower_inverse)
  	return operation_df

  def getResultingSeries(self, reaction_label):
    """
    Return a reaction series, that is, a column
    from a reduced stoichiometry matrix.
    :param str reaction_label:
    :return False/pandas.Series: result_series
    """
    if type(reaction_label) != str:
      return False
    else: 
      if self.mesgraph.rref_df is None:
        result_series = self.mesgraph.echelon_df[reaction_label]
      else:
        result_series = self.mesgraph.rref_df[reaction_label]
      return result_series

  def getOperationStoichiometryMatrix(self, reaction_operations):
  	"""
  	Create a stoichiometry matrix of reactions 
  	related to operations.
  	:param list-ReactionOperation reaction_operations:
  	:return pandas.DataFrame: stoichiometry_df
  	"""
  	reactions = []
  	species = set()
  	for op in reaction_operations:
  	  reaction = self.mesgraph.simple.getReaction(op.reaction)
  	  reactions.append(reaction)
  	  species = species.union({r.molecule.name for r in reaction.reactants})
  	  species = species.union({p.molecule.name for p in reaction.products})
  	operations = np.array([val.operation for val in reaction_operations])
  	stoichiometry_df = self.mesgraph.getStoichiometryMatrix(reactions, list(species))
  	return stoichiometry_df

  def getInferredReaction(self, reaction_operations):
  	"""
  	Create an inferred reaction from reaction_operations.
  	An inferred reaction is a linear combination of reactions.
  	:param list-ReactionOperation reaction_operations:
  	:return SimplifiedReaction: inferred_reaction
  	"""
  	INFERRED_REACTION = "Inferred Reaction"
  	stoichiometry_df = self.getOperationStoichiometryMatrix(reaction_operations)
  	reaction_index = [op.reaction for op in reaction_operations]
  	operation_series = pd.Series([val.operation for val in reaction_operations], index=reaction_index)
  	resulting_reaction = stoichiometry_df.dot(operation_series)
  	# Create a simplified reaction based on the resulting_reaction
  	reactants = []
  	products = []
  	for ms in resulting_reaction.iteritems():
  	  if abs(ms[1]) < TOLERANCE:
  	  	continue
  	  elif ms[1] > 0:
  	  	products.append(MoleculeStoichiometry(molecule=self.mesgraph.simple.getMolecule(ms[0]),
  	  	                                      stoichiometry=ms[1]))
  	  else:
  	  	reactants.append(MoleculeStoichiometry(molecule=self.mesgraph.simple.getMolecule(ms[0]),
  	  	                                      stoichiometry=abs(ms[1])))
  	inferred_reaction = SimplifiedReaction(reactants,
  		                                   products,
  		                                   NULL_STR,
  		                                   self.mesgraph)
  	inferred_reaction.reduceBySOMs()
  	return inferred_reaction

  def reportReactionsInSOM(self, som, reaction_count=0):
    """
    Generate report on reactions created a SOM,
    in order to demonstrate how the SOM was constructed.
    :param som SOM:
    :param int reaction_count:
    :return str: report
    :return int: reaction_count
    """
    report = cn.NULL_STR
    if not som.reactions:
      return report, reaction_count
    reactions = list(som.reactions)
    for r in reactions:
      reaction_count += 1
      report = report + "%d. %s\n" % (reaction_count,
                                      r.makeIdentifier(is_include_kinetics=False))
    return report, reaction_count

  def reportTypeThreeError(self, type_three_errors, explain_details=False):
    """
    Generate a report for Type III errors.
    A Type III error occurs when there is 
    a 1-1 SOMReaction, while there is a preexisting
    MESGraph node between them. 
    :param list-SOMReaction type_three_errors:
    :param bool explain_details:
    :return False/str: report
    :return list-int: error_num
    """
    report = NULL_STR
    error_num = []
    if len(type_three_errors) == 0:
      return report, error_num
    for type3_error in type_three_errors:
      reaction_count = 0
      if type3_error.category != cn.REACTION_1_1:
        print("This canot be a type three error!")
        report = False
        break
      else:
        reaction_label = type3_error.label
        reactant_som = type3_error.reactants[0].som
        product_som = type3_error.products[0].som
        operation_df = self.getOperationMatrix()
        operation_series = operation_df.T[reaction_label]
        reaction_operations = self.convertOperationSeriesToReactionOperations(operation_series)
        # if the number of elements exceeds the threshold, do not explain details
        if len(reaction_operations) > self.explain_threshold:
          explain_details = False
        inferred_reaction = self.getInferredReaction(reaction_operations)
        inferred_som_reaction = self.mesgraph.convertReactionToSOMReaction(inferred_reaction)
        if inferred_som_reaction.getCategory() != cn.REACTION_1_1:
          explain_details = False
        reactant_som = inferred_som_reaction.reactants[0].som
        product_som = inferred_som_reaction.products[0].som
        inequality_reactions = []
        if self.mesgraph.has_edge(reactant_som, product_som):
          inequality_reactions = inequality_reactions + self.mesgraph.get_edge_data(reactant_som, product_som)[cn.REACTION]
        if self.mesgraph.has_edge(product_som, reactant_som):
          inequality_reactions = inequality_reactions + self.mesgraph.get_edge_data(product_som, reactant_som)[cn.REACTION]
        #
        report = report + "We detected a mass imbalance from the following reactions:\n"
        soms = set()
        som_reactions = []
        for r in reaction_operations:
          reaction = self.mesgraph.simple.getReaction(r.reaction)
          reaction_count += 1
          report = report + "\n%d. %s" % (reaction_count, reaction.makeIdentifier(is_include_kinetics=False))
          som_reaction = self.mesgraph.convertReactionToSOMReaction(reaction)
          som_reactions.append(som_reaction)
          soms = soms.union({r.som for r in som_reaction.reactants})
          soms = soms.union({p.som for p in som_reaction.products})
        # calculated the reported number of reactions that constitute a type III error
        error_num.append(reaction_count + len(inequality_reactions))
        if explain_details:
          report = report + "\n\n%s%s\n" % ("-"*NUM_STAR, PARAGRAPH_DIVIDER)
          report = report + "These uni-uni reactions created mass-equivalence.\n" 
          report = report + "(The chemical species within a curly bracket have the same atomic mass.)\n"
        else:
          report = report + "\n"  	  	
        for som in soms:
          if explain_details and som.reactions:
            report = report + "\n%s is inferred by:\n" % som.makeId()
          sub_report, reaction_count = self.reportReactionsInSOM(som, reaction_count)
          report = report + sub_report
        if explain_details:
          report = report + "%s\n" % (PARAGRAPH_DIVIDER)
          report = report + "These multi-uni reactions created mass-inequality.\n\n"  	  
        for r in inequality_reactions:
          reaction = self.mesgraph.simple.getReaction(r)
          reaction_count += 1
          report = report + "%d. %s\n" % (reaction_count, reaction.makeIdentifier(is_include_kinetics=False))
        reaction_count = reaction_count - len(inequality_reactions)
        if explain_details:
          report = report + "%s\n" % (PARAGRAPH_DIVIDER)
          report = report + "Based on the reactions above, we have mass-equivalent pseudo reactions.\n"
          pseudo_reaction_count = 0
          for sr in som_reactions:
            pseudo_reaction_count += 1
            report = report + "\n(pseudo %d.) %s" % (pseudo_reaction_count, sr.identifier)
          report = report +  "\n%s\n" % (PARAGRAPH_DIVIDER)
          report = report + "An operation between pseudo reactions:\n"
          report = report + "\n%.2f * %s" % (reaction_operations[0].operation, reaction_operations[0].reaction)
          for ro in reaction_operations[1:]:
            if ro.operation < 0:
              report = report + " - "
            else:
              report = report + " + "
            report = report + "%.2f * %s\n" % (abs(ro.operation), ro.reaction)
          report = report + "\n\nwill result in a uni-uni reaction:\n"
          report = report + "\n%s\n" % (inferred_som_reaction.identifier)
          report = report + "\n\nmeaning %s and %s have equal mass.\n" % (reactant_som, product_som)
          report = report +  "%s\n" % (PARAGRAPH_DIVIDER)
          report = report + "However, the following mass-equivalent pseudo reaction(s):\n"
          for r in inequality_reactions:
            reaction = self.mesgraph.simple.getReaction(r)
            reaction_count += 1
            report = report + "\n%d. %s" % (reaction_count, reaction.makeIdentifier(is_include_kinetics=False))
            som_reaction = self.mesgraph.convertReactionToSOMReaction(reaction)
            report = report + "\n(pseudo %d.) %s" % (reaction_count, som_reaction.identifier)
          report = report + "\n\nincidates the masses of %s and %s are unequal.\n" % (reactant_som, product_som)
          #
          report = report +  "\n%s%s\n" % (PARAGRAPH_DIVIDER, PARAGRAPH_DIVIDER)
        report = report + "\n%s\n" % (REPORT_DIVIDER) 
    return report, error_num

  def reportEchelonError(self, echelon_errors, explain_details=False):
    """
    Generate a report for echelon errors, i.e.
    mass balance errors from LU decomposition/RREF
    and store in self.report_echelon_errors.
    The operation_df is either mesgraph.lower_inverse
    if LU decomposition yielded echelon_errors,
    and rref_df.dot(lower_inverse) if RREF created errors. 
    :param list-SOMReaction echelon_errors:
    :param bool explain_details:
    :return str: report
    :return list-int: error_num
    """
    report = NULL_STR
    error_num = []
    if len(echelon_errors) == 0:
      return report, error_num
    operation_df = self.getOperationMatrix()
    error_report = NULL_STR
    for reaction in echelon_errors:
      reaction_count = 0
      reaction_label = reaction.label
      operation_series = operation_df.T[reaction_label]
      result_series = self.getResultingSeries(reaction_label)
      reaction_operations = self.convertOperationSeriesToReactionOperations(operation_series)
      # if the number of elements exceeds the threshold, do not explain details
      if len(reaction_operations) > self.explain_threshold:
        explain_details = False
      inferred_reaction = self.getInferredReaction(reaction_operations)
      inferred_som_reaction = self.mesgraph.convertReactionToSOMReaction(inferred_reaction)
      #
      if explain_details:
        report = report + "\nWe detected a mass imbalance\n%s\n" % inferred_reaction.identifier
      else:
        report = report + "\nWe detected a mass imbalance"
      report = report + "\nfrom the following reaction isolation set.\n"
      #
      nonzero_idx = np.array([idx for idx, val in enumerate(result_series) if val != 0.0])
      nonzero_result_series = result_series[nonzero_idx]
      #
      # part 1: reactions that caused mass balance errors
      reaction_operations = self.convertOperationSeriesToReactionOperations(operation_series)
      reported_reactions = [r.reaction for r in reaction_operations]
      reported_som_reactions = []
      for r in reported_reactions:
      	reaction = self.mesgraph.simple.getReaction(r)
      	reaction_count += 1
      	report = report + "\n%d. %s" % (reaction_count, reaction.makeIdentifier(is_include_kinetics=False))
      error_num.append(reaction_count)
      one_side = "--undetermined--"
      if inferred_som_reaction.reactants==[]:
        one_side = "reactant"
      elif inferred_som_reaction.products==[]:
        one_side = "product"
      if one_side == "--undetermined--":
        explain_details = False
      #
      # part 2: SOMs that were canceled by the operation
      canceled_soms = set()
      for r in reported_reactions:
      	sr = self.mesgraph.convertReactionToSOMReaction(self.mesgraph.simple.getReaction(r))
      	reported_som_reactions.append(sr)
      	canceled_soms = canceled_soms.union({r.som for r in sr.reactants})
      	canceled_soms = canceled_soms.union({p.som for p in sr.products})
      canceled_soms = canceled_soms.difference({r.som for r in inferred_som_reaction.reactants})
      canceled_soms = canceled_soms.difference({p.som for p in inferred_som_reaction.products})
      #
      if explain_details:
      	report = report + "\n\n%s%s\n" % ("-"*NUM_STAR, PARAGRAPH_DIVIDER)
      	report = report + "These uni-uni reactions created mass-equivalence.\n"
      	report = report + "(The chemical species within a curly bracket have the same atomic mass.)\n"
      else:
        report = report + "\n"
      for som in canceled_soms:
        if explain_details and som.reactions:
          report = report + "\n%s is inferred by:\n" % som.makeId()
        sub_report, reaction_count = self.reportReactionsInSOM(som, reaction_count)
        report = report + sub_report
      if explain_details:
        report = report + "%s\n" % (PARAGRAPH_DIVIDER)
        report = report + "Based on the uni-uni reactions above, we create mass-equivalent pseudo reactions.\n"
        pseudo_reaction_count = 0
        for sr in reported_som_reactions:
          pseudo_reaction_count += 1
          report = report + "\n(pseudo %d.) %s" % (pseudo_reaction_count, sr.identifier)
        report = report +  "\n%s\n" % (PARAGRAPH_DIVIDER)
        report = report + "An operation between the pseudo reactions:\n"
        report = report + "%.2f * %s" % (reaction_operations[0].operation, reaction_operations[0].reaction)
        for ro in reaction_operations[1:]:
          if ro.operation < 0:
            report = report + " - "
          else:
            report = report + " + "
          report = report + "%.2f * %s" % (abs(ro.operation), ro.reaction)
        # one_side is determined in lines 589-593
        report = report + "\n\nwill result in empty %s with zero mass:\n" % (one_side)
        report = report + "\n%s\n" % (inferred_som_reaction.identifier)
        report = report +  "\n%s%s\n" % (PARAGRAPH_DIVIDER, PARAGRAPH_DIVIDER)
      report = report + "\n%s\n" % (REPORT_DIVIDER)
    return report, error_num

  def reportCancelingError(self, canceling_errors, explain_details=False):
    """
    Generate a report for canceling errors.
    A canceling error occurs when a som_reaction has
    an imbalanced net stoichiometry.
    :param list-SOMReaction:
    :return str: report
    :return list-int: error_num
    """
    report = NULL_STR
    error_num = []
    if len(canceling_errors) == 0:
      return report, error_num
    for error in canceling_errors:
      reaction_count = 0
      label = error.label
      reaction = self.mesgraph.simple.getReaction(label)
      simplified_reaction = SimplifiedReaction(reaction.reactants, reaction.products, label, self.mesgraph)
      simplified_reaction.reduceBySOMs()
      som_reaction = self.mesgraph.convertReactionToSOMReaction(reaction)
      som_reactants = {r.som for r in som_reaction.reactants}
      som_products = {p.som for p in som_reaction.products}
      canceled_soms = list(som_reactants.intersection(som_products))
      reaction_count += 1
      report = report + "\n%d. %s" % (reaction_count, reaction.makeIdentifier(is_include_kinetics=False))
      for som in canceled_soms:
        if len(som.molecules) == 1 and explain_details:
          molecule = list(som.molecules)[0]
          report = report + "\n*%s is a common chemical species in reactants and products, so can be canceled\n" % (molecule.name)
        for reaction in list(som.reactions):
          reaction_count += 1
          reactant = reaction.reactants[0].molecule
          product = reaction.products[0].molecule
          report = report + "\n%d. %s" % (reaction_count, reaction.makeIdentifier(is_include_kinetics=False))
          if explain_details:
            report = report + "\n*%s and %s have the same mass according to the above reaction\n" % (reactant.name, product.name)
      if error.reactants==[]:
        one_side = "reactant"
      elif error.products==[]:
        one_side = "product"
      report = "We detected a mass imbalance\n: %s\n\nfrom the following reaction isolation set:\n" % (simplified_reaction.identifier) + report 
      error_num.append(reaction_count)      
      report = report + "\n%s%s\n" % (PARAGRAPH_DIVIDER, PARAGRAPH_DIVIDER)
    report = report + "\n%s\n" % (REPORT_DIVIDER)
    return report, error_num











from SBMLLint.common import constants as cn
from SBMLLint.common import exceptions
from SBMLLint.common.runner import Runner
from SBMLLint.common.simple_sbml import SimpleSBML
from SBMLLint.tools import sbmllint


import numpy as np
import os
import sys
import unittest


IGNORE_TEST = False
IS_REPORT = False
TEST_FILE = "test_sbmllint.txt"
TEST_OUT_PATH = os.path.join(cn.TEST_DIR, TEST_FILE)
TEST_147_CFG_FILE2 = os.path.join(cn.TEST_DIR,
    "test_BIOMOD147_cfg2.yml")
TEST_147_CFG_FILE = os.path.join(cn.TEST_DIR,
    "test_BIOMOD147_cfg.yml")
TEST_147_SBML_FILE = os.path.join(cn.TEST_DIR,
    "test_BIOMD0000000147_url.xml")


#############################
# Tests
#############################
class TestFunctions(unittest.TestCase):

  def tearDown(self):
    if os.path.isfile(TEST_OUT_PATH):
      os.remove(TEST_OUT_PATH)

  def testLint(self):
    if IGNORE_TEST:
      return
    with open(TEST_OUT_PATH, 'w') as fd:
      result = sbmllint.lint(model_reference=cn.TEST_FILE4, 
          file_out=fd,
          is_report=IS_REPORT,
          mass_balance_check=cn.MOIETY_ANALYSIS)
    self.assertGreaterEqual(
        result.num_reactions, result.num_imbalances)
    with open(TEST_OUT_PATH, 'r') as fd:
      lines = fd.readlines()
    self.assertGreater(len(lines), 0)

  def testLintWithXMLFileFid(self):
    if IGNORE_TEST:
      return
    fid = open(cn.TEST_FILE4, "r")
    with open(TEST_OUT_PATH, 'w') as fd:
      result = sbmllint.lint(model_reference=fid, file_out=fd,
          is_report=IS_REPORT, mass_balance_check=cn.MOIETY_ANALYSIS)
    fid.close()
    self.assertGreater(len(result), 0)

  def testLintWithConfigFid(self):
    if IGNORE_TEST:
      return
    def get(config_fid=None):
      with open(TEST_OUT_PATH, 'w') as fd:
        result = sbmllint.lint(model_reference=cn.TEST_FILE4, file_out=fd,
            config_fid=config_fid,
            is_report=IS_REPORT,
            mass_balance_check=cn.MOIETY_ANALYSIS)
      return result
    #
    result1 = get()
    fid = open(cn.CFG_DEFAULT_PATH)
    result2 = get(config_fid=fid)
    self.assertEqual(result1, result2)
    fid.close()

  def testLint(self):
    if IGNORE_TEST:
      return
    with open(TEST_OUT_PATH, 'w') as fd:
      result = sbmllint.lint(model_reference=cn.TEST_FILE4,
          file_out=fd,
          is_report=True,
          mass_balance_check=cn.MOIETY_ANALYSIS)
    self.assertGreaterEqual(
        result.num_reactions, result.num_imbalances)
    with open(TEST_OUT_PATH, 'r') as fd:
      lines = fd.readlines()
    self.assertGreater(len(lines), 0)

  def testLint2(self):
    if IGNORE_TEST:
      return
    with open(TEST_OUT_PATH, 'w') as fd:
      result = sbmllint.lint(model_reference=cn.TEST_FILE2,
          file_out=fd, is_report=True,
          mass_balance_check=cn.MOIETY_ANALYSIS)
    self.assertGreaterEqual(
        result.num_reactions, result.num_imbalances)
    with open(TEST_OUT_PATH, 'r') as fd:
      lines = fd.readlines()
    self.assertGreater(len(lines), 0)

  def testLint3(self):
    if IGNORE_TEST:
      return
    model = """
    2Glu + 2A_P_P_P -> 2Glu_P + 2A_P_P; 1
    Glu = 0
    A_P_P_P = 0
    Glu_P = 0
    A_P_P = 0
    """
    with open(TEST_OUT_PATH, 'w') as fd:
      try:
        result = sbmllint.lint(model_reference=model, file_out=fd,
            is_report=IS_REPORT,
            mass_balance_check=cn.MOIETY_ANALYSIS)
      except exceptions.MissingTelluriumError:
        return
    self.assertEqual(result.num_reactions, 1)
    self.assertEqual(result.num_imbalances, 0)

  def testLint4(self):
    if IGNORE_TEST:
      return
    model = """
    2Glu_DUMMYMOIETY + 2A__P_3 -> 2Glu_P + 2A_P_P; 1
    Glu_DUMMYMOIETY = 0
    A_P_P = 0
    Glu_P = 0
    A__P_3 = 0
    A = 0
    """
    with open(TEST_OUT_PATH, 'w') as fd:
      try:
        result = sbmllint.lint(model_reference=model, file_out=fd,
            is_report=IS_REPORT,
            mass_balance_check=cn.MOIETY_ANALYSIS)
      except exceptions.MissingTelluriumError:
        return
    self.assertEqual(result.num_reactions, 1)
    self.assertEqual(result.num_imbalances, 0)

  def testLint5(self):
    if IGNORE_TEST:
      return
    model = """
    A -> B; 1
    B -> B + DUMMYMOLECULE;1
    A = 0
    B = 0
    DUMMYMOLECULE = 0
    """
    with open(TEST_OUT_PATH, 'w') as fd:
      try:
        result = sbmllint.lint(model_reference=model,
                               file_out=fd,
                               mass_balance_check="games",
                               implicit_games=False,
                               is_report=IS_REPORT
                               )
      except exceptions.MissingTelluriumError:
        return
    self.assertTrue(result)
    with open(TEST_OUT_PATH, 'w') as fd:
      result = sbmllint.lint(model_reference=model,
                             file_out=fd,
                             mass_balance_check="games",
                             implicit_games=True,
                             is_report=IS_REPORT
                             )    
    self.assertFalse(result)

  def testRemoveIgnored(self):
    if IGNORE_TEST:
      return
    implicit = "MA"
    path = os.path.join(cn.BIOMODELS_DIR, cn.TEST_FILE13)
    simple = SimpleSBML()
    simple.initialize(path)
    simple = sbmllint.removeIgnored(simple, implicit)
    implicit_reactions = []
    for r in simple.reactions:
      reactants = [reactant.molecule.name for reactant in r.reactants]
      products = [product.molecule.name for product in r.products]
      if (implicit in reactants) or (implicit in products):
        implicit_reactions.append(r.label)
    self.assertTrue(len(implicit_reactions) == 0)

  def testMain(self):
    if IGNORE_TEST:
      return
    return
    # FIXME: This test fails in traves
    module_dir = os.path.abspath(os.curdir)
    for ele in ["SBMLLint", "tools"]:
      module_dir = os.path.join(module_dir, ele)
    module_path = os.path.join(module_dir, "sbmllint.py")
    runner = Runner(module_path)
    runner.execute([cn.TEST_FILE4], '')
    self.assertGreater(runner.output.count('\n'), 0)

  def testBIOMOD147(self):
    if IGNORE_TEST:
      return
    with open(TEST_OUT_PATH, 'w') as fd:
      result = sbmllint.lint(model_reference=TEST_147_SBML_FILE,
          config_fid=open(TEST_147_CFG_FILE, "r"),
          file_out=fd,
          mass_balance_check=cn.MOIETY_ANALYSIS,
          implicit_games=False,
          is_report=IS_REPORT,
          )
    # Verify that exclicit declaration is decomposed
    # into moieties
    molecule_name = "IkBeIKKNFkB"
    self.assertFalse("%s:" % molecule_name in result.report)

  def testBIOMOD147_2(self):
    if IGNORE_TEST:
      return
    with open(TEST_OUT_PATH, 'w') as fd:
      result = sbmllint.lint(model_reference=TEST_147_SBML_FILE,
          config_fid=open(TEST_147_CFG_FILE2, "r"),
          file_out=fd,
          mass_balance_check=cn.MOIETY_ANALYSIS,
          implicit_games=False,
          is_report=IS_REPORT,
          )
    # Verify that exclicit declaration is decomposed
    # into moieties
    self.assertFalse("cytoplasm:" in result.report)
    self.assertFalse("nulceus:" in result.report)


if __name__ == '__main__':
  unittest.main()



#!/usr/bin/env python
"""
Performs Linear Programming Analysis to detection stoichiometric
inconsistencies.
"""

from SBMLLint.common import constants as cn
from SBMLLint.common import simple_sbml
from SBMLLint.common import stoichiometry_matrix
from SBMLLint.common import util

import argparse
import sys


def LPAnalysis(fid, is_report=False):
  """
  Does LP analysis for a simple model.
  :param IOStream fid: XML file
  :param bool: True if model is stoichiometric consistent.
  """
  simple = simple_sbml.SimpleSBML()
  simple.initialize(fid)
  sm_matrix = stoichiometry_matrix.StoichiometryMatrix(
      simple=simple)
  is_consistent = sm_matrix.isConsistent(is_report_warning=is_report)
  if is_consistent:
    print("Model is consistent.")
  else:
    print("Model is NOT consistent!")
  return is_consistent

def main():
  def str2Bool(stg):
    if "T" in stg.upper():
      return True
    if "F" in stg.upper():
      return False
    return False
  #
  parser = argparse.ArgumentParser(
      description='LP Analysis of SBML file(s).')
  parser.add_argument('xml_fid', type=open,
      help='SBML file or zipfile')
  parser.add_argument('--report_warnings', nargs=1,
      type=str2Bool,
      help="Print warnings if ill-formed matrix True or False",
      default = ['True'])
  args = parser.parse_args()
  for fid in util.getNextFid(args.xml_fid):
    util.runFunction(LPAnalysis,
        pargs=[fid], 
        kwargs={"is_report": args.report_warnings[0]},
        )


if __name__ == '__main__':
  main()



"""
Test for GAMES Plus (Reduced) Row Echelon Form (GAMES_PP)
"""
from SBMLLint.common import constants as cn
from SBMLLint.common.molecule import Molecule
from SBMLLint.common.reaction import Reaction
from SBMLLint.common.simple_sbml import SimpleSBML
from SBMLLint.games.games_pp import SOMStoichiometry, SOMReaction, GAMES_PP
from SBMLLint.games.som import SOM
from SBMLLint.common import simple_sbml

import numpy as np
import pandas as pd
import os
import re
import unittest

IGNORE_TEST = False

ZERO = 0
ONE = 1
ZERO_F = 0.0
ONE_F = 1.0
# BIOMD0000000383
# Molecule names
PGA = "PGA"
RUBP = "RuBP"
CO2 = "CO2"
# Reaction names
PGA_CONS = "PGA_cons"
PGA_CONS_SOMREACTION_IDENTIFIER = "PGA_cons: {PGA} -> {RuBP}"
PGA_PROD_VC = "PGA_prod_Vc"
PGA_PROD_VO = "PGA_prod_Vo"
# SOMStoichiometry identifier
RUBP_ONE = "{RuBP} * 1.00"
# SOMStoichiometry NADPH identifier and stoichioemtry
SOM_NADPH = "{NADPH}"
SOM_PGA = "{PGA}"
SOM_RUBP = "{RuBP}"
NADPH_STOICHIOMETRY = 2.0
# StoichiometryMatrix value
PGA_CONS_WITH_PGA = -1.0
PGA_PROD_VC_WITH_RUBP = -1.0
#"
REACTION = "reaction"
NUM_REACTIONS = 4
NUM_MOLECULES = 5
#
# BIOMD0000000018
# Reactions
CH2FH4toHCHO = "CH2FH4toHCHO"
HCHOtoCH2FH4 = "HCHOtoCH2FH4"
# Molecules
CH2FH4 = "CH2FH4"
FH4 = "FH4"
HCHO = "HCHO"

#############################
# Tests
#############################
class TestSOMStoichiometry(unittest.TestCase):

  def setUp(self):
    self.simple = SimpleSBML()
    self.simple.initialize(cn.TEST_FILE_GAMES_PP1)
    self.reaction = self.simple.getReaction(PGA_CONS)
    self.rubp = self.reaction.products[0]
    self.rubl_ss = SOMStoichiometry(
        som = SOM([self.rubp.molecule]),
        stoichiometry = self.rubp.stoichiometry
        )

  def testConstructor(self):
    if IGNORE_TEST:
      return
    self.assertTrue(isinstance(self.rubl_ss, SOMStoichiometry))
    self.assertTrue(isinstance(self.rubl_ss.som, SOM))
    self.assertTrue(isinstance(self.rubl_ss.stoichiometry, float))

  def testMakeId(self):
  	self.assertEqual(self.rubl_ss.identifier, RUBP_ONE)


class TestSOMReaction(unittest.TestCase):

  def setUp(self):
    self.simple = SimpleSBML()
    self.simple.initialize(cn.TEST_FILE_GAMES_PP1)
    self.reaction = self.simple.getReaction(PGA_CONS)
    self.pga = self.reaction.reactants[0]
    self.pga_ss = SOMStoichiometry(
        som = SOM([self.pga.molecule]),
        stoichiometry = self.pga.stoichiometry
        )
    self.rubp = self.reaction.products[0]
    self.rubp_ss = SOMStoichiometry(
        som = SOM([self.rubp.molecule]),
        stoichiometry = self.rubp.stoichiometry
        )
    self.som_reaction = SOMReaction(
        reactants=[self.pga_ss],
        products=[self.rubp_ss],
        label=self.reaction.label)

  def testConstructor(self):
    self.assertTrue(isinstance(self.reaction, Reaction))
    self.assertEqual(self.pga.molecule.name, PGA)
    self.assertEqual(self.rubp.molecule.name, RUBP)    
    self.assertTrue(isinstance(self.pga_ss, SOMStoichiometry))
    self.assertTrue(isinstance(self.rubp_ss, SOMStoichiometry))
    self.assertTrue(isinstance(self.som_reaction, SOMReaction))      

  def testMakeId(self):
  	self.assertEqual(
  	    self.som_reaction.makeId(),
  	    PGA_CONS_SOMREACTION_IDENTIFIER
  	    )

  def testGetCategory(self):
  	self.assertEqual(
  	    self.som_reaction.category,
  	    cn.REACTION_1_1)


class TestGAMES_PP(unittest.TestCase):

  def setUp(self):
    # BIOMD0000000383
    self.simple1 = SimpleSBML()
    self.simple1.initialize(cn.TEST_FILE_GAMES_PP1)
    self.games_pp = GAMES_PP(self.simple1)
    # BIOMD0000000018
    self.simple2 = SimpleSBML()
    self.simple2.initialize(cn.TEST_FILE_GAMES_PP2)

  def testConstructor(self):
    if IGNORE_TEST:
      return
    self.assertEqual(len(self.games_pp.reactions), NUM_REACTIONS)
    for r in self.games_pp.reactions:
      print(r.category)
      self.assertFalse(r.category==cn.REACTION_BOUNDARY)
    self.assertEqual(len(self.games_pp.molecules), NUM_MOLECULES)
    for m in self.games_pp.molecules:
      self.assertTrue(isinstance(m, Molecule))
    self.assertEqual(len(self.games_pp.soms), NUM_MOLECULES)
    self.assertTrue(isinstance(self.games_pp.soms[0], SOM))
    init_identifier = ""
    for som in self.games_pp.soms:
      init_identifier = init_identifier + som.identifier
      if som != self.games_pp.soms[-1]:
        init_identifier = init_identifier + ";"
    self.assertEqual(self.games_pp.identifier, init_identifier)

  def testConvertReactionToSOMReaction(self):
    if IGNORE_TEST:
      return
    reaction1 = self.games_pp.simple.getReaction(PGA_CONS)
    reaction2 = self.games_pp.simple.getReaction(PGA_PROD_VC)
    som_reaction1 = self.games_pp.convertReactionToSOMReaction(reaction1)
    som_reaction2 = self.games_pp.convertReactionToSOMReaction(reaction2)
    self.assertTrue(isinstance(som_reaction1, SOMReaction))
    self.assertTrue(isinstance(som_reaction2, SOMReaction))
    ms_rubp = som_reaction1.products[0]
    self.assertTrue(isinstance(ms_rubp, SOMStoichiometry))
    self.assertEqual(list(ms_rubp.som.molecules)[0], 
        self.games_pp.simple.getMolecule(RUBP))
    ms_nadph = None
    for ms in som_reaction2.reactants:
      if ms.som.identifier == SOM_NADPH:
        ms_nadph = ms
        break
    self.assertTrue(ms_nadph.som.identifier==SOM_NADPH)
    self.assertEqual(ms_nadph.stoichiometry, NADPH_STOICHIOMETRY)

  def testGetStoichiometryMatrix(self):
    if IGNORE_TEST:
      return
    # For regular stoichiometry matrix
    mat = self.games_pp.getStoichiometryMatrix(
        self.games_pp.reactions,
        self.games_pp.molecules,
        som=False)
    self.assertTrue(isinstance(mat, pd.DataFrame))
    self.assertEqual(mat.shape, (NUM_MOLECULES, NUM_REACTIONS))
    self.assertEqual(mat[PGA_CONS][PGA], PGA_CONS_WITH_PGA)
    self.assertEqual(mat[PGA_PROD_VC][RUBP], PGA_PROD_VC_WITH_RUBP)
    # For SOM stoichiometry matrix
    som_reactions = []
    for r in self.games_pp.reactions:
      som_reactions.append(self.games_pp.convertReactionToSOMReaction(r))
    som_mat = self.games_pp.getStoichiometryMatrix(
        som_reactions,
        self.games_pp.nodes,
        som=True)
    self.assertTrue(isinstance(som_mat, pd.DataFrame))
    self.assertEqual(som_mat[PGA_CONS][SOM_PGA], PGA_CONS_WITH_PGA)
    self.assertEqual(som_mat[PGA_PROD_VC][SOM_RUBP], PGA_PROD_VC_WITH_RUBP)

  def testDecomposeMatrix(self):
    if IGNORE_TEST:
      return
    # should be all None before decomposition
    self.assertTrue(self.games_pp.perm_inverse is None)
    self.assertTrue(self.games_pp.permuted_matrix is None)
    self.assertTrue(self.games_pp.lower_inverse is None)
    self.assertTrue(self.games_pp.echelon_df is None)
    som_reactions = []
    for r in self.games_pp.reactions:
      som_reactions.append(self.games_pp.convertReactionToSOMReaction(r))
    som_mat = self.games_pp.getStoichiometryMatrix(
        som_reactions,
        self.games_pp.nodes,
        som=True)
    echelon = self.games_pp.decomposeMatrix(som_mat).T
    self.assertFalse(self.games_pp.perm_inverse is None)
    self.assertFalse(self.games_pp.permuted_matrix is None)
    self.assertFalse(self.games_pp.lower_inverse is None)
    self.assertFalse(self.games_pp.echelon_df is None)
    self.assertTrue(isinstance(self.games_pp.perm_inverse, np.ndarray))    
    self.assertTrue(isinstance(self.games_pp.permuted_matrix, pd.DataFrame))
    self.assertTrue(isinstance(self.games_pp.lower_inverse, pd.DataFrame))
    self.assertTrue(isinstance(self.games_pp.echelon_df, pd.DataFrame))
    # Checking lower echelon - lower left element should be 0.0 (zero)
    self.assertTrue(echelon.iloc[echelon.shape[0]-1][0]==ZERO_F)
    # On the ohter hand, upper left should be nonzero
    self.assertFalse(echelon.iloc[0][0]==ZERO_F)

  def testGetRREFMatrix(self):
    if IGNORE_TEST:
      return
    self.assertTrue(self.games_pp.rref_operation is None) 
    self.assertTrue(self.games_pp.rref_df is None)
    som_reactions = []
    for r in self.games_pp.reactions:
      som_reactions.append(self.games_pp.convertReactionToSOMReaction(r))
    som_mat = self.games_pp.getStoichiometryMatrix(
        som_reactions,
        self.games_pp.nodes,
        som=True)
    echelon = self.games_pp.decomposeMatrix(som_mat).T
    rref = self.games_pp.getRREFMatrix(echelon)
    self.assertTrue(isinstance(self.games_pp.rref_operation, pd.DataFrame))
    self.assertTrue(isinstance(self.games_pp.rref_df, pd.DataFrame))
    # Choose the last row of rref.T 
    last_row = rref.T.iloc[-1:]
    # Get the first nonzero index
    nonzero_species = (last_row != 0).idxmax(axis=1)[0]
    # The SUM of nonzero species column must be the same as the single value
    self.assertEqual(last_row[nonzero_species][0], sum(rref.T[nonzero_species]))

  def testConverMatrixToSOMReactions(self):
    if IGNORE_TEST:
      return
    som_reactions1 = []
    for r in self.games_pp.reactions:
      som_reactions1.append(self.games_pp.convertReactionToSOMReaction(r))
    som_mat = self.games_pp.getStoichiometryMatrix(
        som_reactions1,
        self.games_pp.nodes,
        som=True)
    som_reactions2 = self.games_pp.convertMatrixToSOMReactions(som_mat)
    sr1 = None
    sr2 = None
    for sr in som_reactions1:
      if sr.label == PGA_PROD_VC:
        sr1 = sr
        break
    for sr in som_reactions2:
      if sr.label == PGA_PROD_VC:
        sr2 = sr
        break
    sr1_reactants = {ms.som for ms in sr1.reactants}
    sr2_reactants = {ms.som for ms in sr2.reactants}
    sr1_products = {ms.som for ms in sr1.products}
    sr2_products = {ms.som for ms in sr2.products}
    sr1_reactsom = sum([ms.stoichiometry for ms in sr1.reactants])
    sr2_reactsom = sum([ms.stoichiometry for ms in sr2.reactants])
    sr1_prodsom = sum([ms.stoichiometry for ms in sr1.products])
    sr2_prodsom = sum([ms.stoichiometry for ms in sr2.products])
    self.assertEqual(sr1_reactants, sr2_reactants)
    self.assertEqual(sr1_products, sr2_products)
    self.assertEqual(sr1_reactsom, sr2_reactsom)
    self.assertEqual(sr1_prodsom, sr2_prodsom)

  def testGetNode(self):
    if IGNORE_TEST:
      return
    co2 = self.games_pp.simple.getMolecule(CO2)
    co2_node = self.games_pp.getNode(co2)
    self.assertEqual(type(co2_node), SOM)
    self.assertEqual(co2_node.molecules, {co2})

  def testMergeNodes(self):
    if IGNORE_TEST:
      return
    reaction = self.games_pp.simple.getReaction(PGA_CONS)
    pga = self.games_pp.simple.getMolecule(PGA)
    rubp = self.games_pp.simple.getMolecule(RUBP)
    som_pga = self.games_pp.getNode(pga)
    som_rubp = self.games_pp.getNode(rubp)
    som_pga_rubp = self.games_pp.mergeNodes(som_pga, som_rubp, reaction)
    self.assertTrue(isinstance(som_pga_rubp, SOM))
    self.assertTrue(pga in som_pga_rubp.molecules)
    self.assertTrue(rubp in som_pga_rubp.molecules)
    self.assertTrue(reaction in som_pga_rubp.reactions)

  def testAddReaction(self):
    if IGNORE_TEST:
      return
    self.assertEqual(len(self.games_pp.reactions_lu), 0)
    reaction = self.games_pp.simple.getReaction(PGA_CONS)
    self.games_pp.addReaction(reaction)
    self.assertTrue(reaction in self.games_pp.reactions_lu)
    self.games_pp.addReaction(reaction)
    self.assertTrue(len(self.games_pp.reactions_lu), 1)

  def testProcessUniUniReaction(self):
    if IGNORE_TEST:
      return
    reaction = self.games_pp.simple.getReaction(PGA_CONS)
    som = self.games_pp.processUniUniReaction(reaction)
    self.assertTrue(isinstance(som, SOM))
    self.assertTrue(som in self.games_pp.nodes)
    pga = self.games_pp.simple.getMolecule(PGA)
    rubp = self.games_pp.simple.getMolecule(RUBP)
    self.assertTrue(pga in som.molecules)
    self.assertTrue(rubp in som.molecules)

  def testAddArc(self):
    if IGNORE_TEST:
      return
    reaction = self.games_pp.simple.getReaction(PGA_PROD_VC)
    co2 = self.games_pp.simple.getMolecule(CO2)
    pga = self.games_pp.simple.getMolecule(PGA)
    som_co2 = self.games_pp.getNode(co2)
    som_pga = self.games_pp.getNode(pga)
    self.games_pp.addArc(som_co2, som_pga, reaction)
    self.assertTrue(self.games_pp.has_edge(som_co2, som_pga))
    self.assertEqual(self.games_pp.get_edge_data(som_co2, som_pga)[REACTION], [PGA_PROD_VC])

  def testProcessUniMultiReaction(self):
    if IGNORE_TEST:
      return
    games_pp2 = GAMES_PP(self.simple2)
    self.assertTrue(isinstance(games_pp2, GAMES_PP))
    reaction = games_pp2.simple.getReaction(CH2FH4toHCHO)
    reactant1 = games_pp2.simple.getMolecule(CH2FH4)
    product1 = games_pp2.simple.getMolecule(FH4)
    product2 = games_pp2.simple.getMolecule(HCHO)
    games_pp2.processUniMultiReaction(reaction)
    som_reactant1 = games_pp2.getNode(reactant1)
    som_product1 = games_pp2.getNode(product1)
    som_product2 = games_pp2.getNode(product2)
    self.assertTrue(games_pp2.has_edge(som_product1, som_reactant1))
    self.assertTrue(games_pp2.has_edge(som_product2, som_reactant1))

  def testProcessMultiUniReaction(self):
    if IGNORE_TEST:
      return
    games_pp2 = GAMES_PP(self.simple2)
    self.assertTrue(isinstance(games_pp2, GAMES_PP))
    reaction = games_pp2.simple.getReaction(HCHOtoCH2FH4)
    reactant1 = games_pp2.simple.getMolecule(FH4)
    reactant2 = games_pp2.simple.getMolecule(HCHO)
    product1 = games_pp2.simple.getMolecule(CH2FH4)
    games_pp2.processMultiUniReaction(reaction)
    som_product1 = games_pp2.getNode(product1)
    som_reactant1 = games_pp2.getNode(reactant1)
    som_reactant2 = games_pp2.getNode(reactant2)
    self.assertTrue(games_pp2.has_edge(som_reactant1, som_product1))
    self.assertTrue(games_pp2.has_edge(som_reactant2, som_product1))

  def testProcessEqualSOMReaction(self):
    if IGNORE_TEST:
      return
    print("type three", self.games_pp.type_three_errors)
    self.assertEqual(len(self.games_pp.type_three_errors), ZERO)
    # Add an arc
    reaction = self.games_pp.simple.getReaction(PGA_PROD_VC)
    co2 = self.games_pp.simple.getMolecule(CO2)
    pga = self.games_pp.simple.getMolecule(PGA)
    som_co2 = self.games_pp.getNode(co2)
    som_pga = self.games_pp.getNode(pga)
    self.games_pp.addArc(som_co2, som_pga, reaction)
    # Process a dummy reaction
    soms_co2 = SOMStoichiometry(som_co2, 1.0)
    soms_pga = SOMStoichiometry(som_pga, 1.0)
    som_reaction = SOMReaction([soms_co2], [soms_pga], "dummy")
    self.games_pp.processEqualSOMReaction(som_reaction)
    self.assertTrue(len(self.games_pp.type_three_errors) > ZERO)

  def testProcessUnequalSOMReaction(self):
    if IGNORE_TEST:
      return
    games_pp2 = GAMES_PP(self.simple2)
    self.assertTrue(len(games_pp2.edges)==0)
    reaction = games_pp2.simple.getReaction(CH2FH4toHCHO)
    som_reaction = games_pp2.convertReactionToSOMReaction(reaction)
    games_pp2.processUnequalSOMReaction(som_reaction)
    self.assertTrue(len(games_pp2.edges)==2)
    reactant1 = games_pp2.simple.getMolecule(FH4)
    reactant2 = games_pp2.simple.getMolecule(HCHO)
    product1 = games_pp2.simple.getMolecule(CH2FH4)
    games_pp2.processUniMultiReaction(reaction)
    som_product1 = games_pp2.getNode(product1)
    som_reactant1 = games_pp2.getNode(reactant1)
    som_reactant2 = games_pp2.getNode(reactant2)
    self.assertTrue(games_pp2.has_edge(som_reactant1, som_product1))
    self.assertTrue(games_pp2.has_edge(som_reactant2, som_product1))

  def testAddTypeOneError(self):
    if IGNORE_TEST:
      return
    games_pp2 = GAMES_PP(self.simple2)
    self.assertTrue(len(games_pp2.type_one_errors)==ZERO)
    reaction = games_pp2.simple.getReaction(CH2FH4toHCHO)
    reactant1 = games_pp2.simple.getMolecule(FH4)
    product1 = games_pp2.simple.getMolecule(CH2FH4)
    games_pp2.addTypeOneError(reactant1, product1, reaction)
    self.assertTrue(len(games_pp2.type_one_errors)==ONE)
    error = games_pp2.type_one_errors[ZERO]
    self.assertEqual(error.node1, reactant1.name)
    self.assertEqual(error.node2, product1.name)
    self.assertEqual(error.reactions, [reaction.label])

  def testChekcTypeOneError(self):
    if IGNORE_TEST:
      return
    games_pp1 = GAMES_PP(self.simple1)
    reaction = games_pp1.simple.getReaction(PGA_CONS)
    pga = games_pp1.simple.getMolecule(PGA)
    rubp = games_pp1.simple.getMolecule(RUBP)
    som_pga = games_pp1.getNode(pga)
    som_rubp = games_pp1.getNode(rubp)
    som_pga_rubp = games_pp1.mergeNodes(som_pga, som_rubp, reaction)
    self.assertTrue(len(games_pp1.type_one_errors)==ZERO)
    is_error = games_pp1.checkTypeOneError((pga, rubp), reaction)
    self.assertTrue(is_error)
    error = games_pp1.type_one_errors[ZERO]
    self.assertEqual(error.node1, pga.name)
    self.assertEqual(error.node2, rubp.name)
    self.assertEqual(error.reactions, [reaction.label])

  def testCheckTypeTwoError(self):
    if IGNORE_TEST:
      return    
    # Create a dummy cycle by adding two conflicting arcs
    games_pp2 = GAMES_PP(self.simple2)
    self.assertTrue(isinstance(games_pp2, GAMES_PP))
    unimulti_reaction = games_pp2.simple.getReaction(CH2FH4toHCHO)
    multiuni_reaction = games_pp2.simple.getReaction(HCHOtoCH2FH4)
    fh4 = games_pp2.simple.getMolecule(FH4)
    # hcho = games_pp2.simple.getMolecule(HCHO)
    ch2fh4 = games_pp2.simple.getMolecule(CH2FH4)
    som_fh4 = games_pp2.getNode(fh4)
    som_ch2fh4 = games_pp2.getNode(ch2fh4)
    # do we need the next two methods if we're giving a cycle? 
    games_pp2.addArc(som_fh4, som_ch2fh4, unimulti_reaction)
    games_pp2.addArc(som_ch2fh4, som_fh4, multiuni_reaction) 
    self.assertTrue(len(games_pp2.type_two_errors)==ZERO)
    games_pp2.checkTypeTwoError()
    self.assertTrue(len(games_pp2.type_two_errors)==ONE)
    error = games_pp2.type_two_errors[ZERO]
    self.assertTrue(games_pp2.has_edge(error[ZERO], error[ONE]))
    self.assertTrue(games_pp2.has_edge(error[ONE], error[ZERO]))
    error_reactions = set(games_pp2.get_edge_data(error[ZERO], error[ONE])[REACTION])
    error_reactions = error_reactions.union(set(games_pp2.get_edge_data(error[ONE], error[ZERO])[REACTION]))
    self.assertTrue(CH2FH4toHCHO in error_reactions)
    self.assertTrue(HCHOtoCH2FH4 in error_reactions)

  def testProcessErrorReaction(self):
    if IGNORE_TEST:
      return  
    games_pp1 = GAMES_PP(self.simple1)
    self.assertTrue(len(games_pp1.echelon_errors)==ZERO)
    reaction = self.games_pp.simple.getReaction(PGA_PROD_VO)
    games_pp1.processErrorReaction(reaction)
    self.assertTrue(len(games_pp1.echelon_errors)==ONE)
    self.assertTrue(reaction in games_pp1.echelon_errors)

  def testAnalyze(self):
    if IGNORE_TEST:
      return  
    games_pp1 = GAMES_PP(self.simple1)
    games_pp2 = GAMES_PP(self.simple2)
    self.assertTrue(games_pp1.analyze(rref=False))
    self.assertTrue(games_pp2.analyze())
    self.assertTrue(len(games_pp1.echelon_errors)>ZERO)
    self.assertTrue(len(games_pp1.type_one_errors)==ZERO)
    self.assertTrue(len(games_pp1.type_two_errors)==ZERO)
    self.assertTrue(len(games_pp2.type_one_errors)>ZERO)

if __name__ == '__main__':
  unittest.main()
























































from SBMLLint.common import constants as cn
from SBMLLint.common.runner import Runner
from SBMLLint.tools import print_reactions


import numpy as np
import os
import sys
import unittest


IGNORE_TEST = False
TEST_FILE = "test_print_reactions.txt"
TEST_OUT_PATH = os.path.join(cn.TEST_DIR, TEST_FILE)


#############################
# Tests
#############################
class TestFunctions(unittest.TestCase):

  def tearDown(self):
    if os.path.isfile(TEST_OUT_PATH):
      os.remove(TEST_OUT_PATH)

  def testPrettyPrint(self):
    with open(TEST_OUT_PATH, 'w') as fd:
      print_reactions.prettyPrint(cn.TEST_FILE, file_out=fd)
    with open(TEST_OUT_PATH, 'r') as fd:
      lines = fd.readlines()
    self.assertEqual(len(lines), cn.NUM_REACTIONS)

  def testMain(self):
    return
    # FIXME: This test fails in traves
    module_dir = os.path.abspath(os.curdir)
    for ele in ["SBMLLint", "tools"]:
      module_dir = os.path.join(module_dir, ele)
    module_path = os.path.join(module_dir, "print_reactions.py")
    runner = Runner(module_path)
    runner.execute([cn.TEST_FILE], '')
    self.assertEqual(runner.output.count('\n'), cn.NUM_REACTIONS)


if __name__ == '__main__':
  unittest.main()
